/* 
   A C-program for MT19937, with initialization improved 2002/1/26.
   Coded by Takuji Nishimura and Makoto Matsumoto.

   Before using, initialize the state by using init_genrand(seed)  
   or init_by_array(init_key, key_length).

   Copyright (C) 1997 - 2002, Makoto Matsumoto and Takuji Nishimura,
   All rights reserved.                          

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

     1. Redistributions of source code must retain the above copyright
        notice, this list of conditions and the following disclaimer.

     2. Redistributions in binary form must reproduce the above copyright
        notice, this list of conditions and the following disclaimer in the
        documentation and/or other materials provided with the distribution.

     3. The names of its contributors may not be used to endorse or promote 
        products derived from this software without specific prior written 
        permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


   Any feedback is very welcome.
   http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/emt.html
   email: m-mat @ math.sci.hiroshima-u.ac.jp (remove space)
*/

#include "m_random.h"

/* Period parameters */  
#define N 624
#define M 397

#define MATRIX_A 0x9908b0dfUL   /* constant vector a */
#define UPPER_MASK 0x80000000UL /* most significant w-r bits */
#define LOWER_MASK 0x7fffffffUL /* least significant r bits */

static unsigned long mt[N]; /* the array for the state vector  */
static int mti = N + 1; /* mti==N+1 means mt[N] is not initialized */

/* -------------------------------------------------------------------------- */

public void random_init(unsigned long s)
{
    /** @brief initializes state[N] with a seed */

    mt[0]= s & 0xffffffffUL;

    for (mti = 1; mti < N; mti ++) {
        mt[mti] = 
        (1812433253UL * (mt[mti-1] ^ (mt[mti-1] >> 30)) + mti); 
        /* See Knuth TAOCP Vol2. 3rd Ed. P.106 for multiplier. */
        /* In the previous versions, MSBs of the seed affect   */
        /* only MSBs of the array mt[].                        */
        /* 2002/01/09 modified by Makoto Matsumoto             */
        mt[mti] &= 0xffffffffUL;
        /* for >32 bit machines */
    }
}

/* -------------------------------------------------------------------------- */

public m_random *random_initr(unsigned long s)
{
    /** @brief initializes state[N] with a seed (thread safe) */

    m_random *ret = NULL;
    int i = 1;

    if (! (ret = malloc(sizeof(*ret))) ) {
        perror(ERR(random_initr, malloc));
        return NULL;
    }

    if (! (ret->_mt = malloc(N * sizeof(*ret->_mt))) ) {
        perror(ERR(random_initr, malloc));
        free(ret); return NULL;
    }

    ret->_mt[0]= s & 0xffffffffUL;

    for (i = 1; i < N; i ++) {
        ret->_mt[i] =
        (1812433253UL * (ret->_mt[i - 1] ^ (ret->_mt[i - 1] >> 30)) + i);
        ret->_mt[i] &= 0xffffffffUL;
    }

    ret->_mti = i;

    return ret;
}

/* -------------------------------------------------------------------------- */

public inline void random_finir(m_random *state)
{
    /** @brief destroys the state (thread safe) */

    free(state->_mt); free(state);
}

/* -------------------------------------------------------------------------- */

public void random_arrayinit(const unsigned long init_key[],
                             unsigned long key_length       )
{
    /** @brief initialize by an array with array-length */

    unsigned int i = 1, j = 0, k = 0;

    random_init(19650218UL);

    k = (N > key_length ? N : key_length);

    for (; k; k --) {
        mt[i] = (mt[i] ^ ((mt[i - 1] ^ (mt[i - 1] >> 30)) * 1664525UL))
          + init_key[j] + j; /* non linear */
        mt[i] &= 0xffffffffUL; /* for WORDSIZE > 32 machines */
        i ++; j ++;
        if (i >= N) { mt[0] = mt[N - 1]; i = 1; }
        if (j >= key_length) j = 0;
    }
    
    for (k = N - 1; k; k --) {
        mt[i] = (mt[i] ^ ((mt[i - 1] ^ (mt[i - 1] >> 30)) * 1566083941UL))
          - i; /* non linear */
        mt[i] &= 0xffffffffUL; /* for WORDSIZE > 32 machines */
        i ++;
        if (i >= N) { mt[0] = mt[N - 1]; i = 1; }
    }

    mt[0] = 0x80000000UL; /* MSB is 1; assuring non-zero initial array */
}

/* -------------------------------------------------------------------------- */

public m_random *random_arrayinitr(const unsigned long init_key[],
                                   unsigned long key_length       )
{
    /** @brief initialize by an array with array-length (thread safe) */

    unsigned int i = 1, j = 0, k = 0;
    m_random *ret = NULL;

    ret = random_initr(19650218UL);

    k = (N > key_length ? N : key_length);

    for (; k; k --) {
        ret->_mt[i] =
        ( ret->_mt[i] ^
          ((ret->_mt[i - 1] ^ (ret->_mt[i - 1] >> 30)) * 1664525UL)
        ) + init_key[j] + j; /* non linear */

        ret->_mt[i] &= 0xffffffffUL; /* for WORDSIZE > 32 machines */

        i ++; j ++;
        if (i >= N) { ret->_mt[0] = ret->_mt[N - 1]; i = 1; }
        if (j >= key_length) j = 0;
    }
    
    for (k = N - 1; k; k --) {
        ret->_mt[i] =
        ( ret->_mt[i] ^
          ((ret->_mt[i - 1] ^ (ret->_mt[i - 1] >> 30)) * 1566083941UL)
        ) - i; /* non linear */

        ret->_mt[i] &= 0xffffffffUL; /* for WORDSIZE > 32 machines */

        i ++;
        if (i >= N) { ret->_mt[0] = ret->_mt[N - 1]; i = 1; }
    }

    ret->_mt[0] = 0x80000000UL; /* MSB is 1; assuring non-zero initial array */

    return ret;
}

/* -------------------------------------------------------------------------- */

public unsigned long random_int32(void)
{
    /** @brief generates a random number on [0,0xffffffff]-interval */

    unsigned long y;
    static unsigned long mag01[2] = {0x0UL, MATRIX_A};
    /* mag01[x] = x * MATRIX_A  for x=0,1 */

    if (mti >= N) { /* generate N words at one time */
        int kk;

        if (mti == N + 1)   /* if random_init() has not been called, */
            random_init(5489UL); /* a default initial seed is used */

        for (kk = 0; kk < N - M; kk ++) {
            y = (mt[kk] & UPPER_MASK) | (mt[kk + 1] & LOWER_MASK);
            mt[kk] = mt[kk + M] ^ (y >> 1) ^ mag01[y & 0x1UL];
        }

        for ( ; kk < N - 1; kk ++) {
            y = (mt[kk] & UPPER_MASK) | (mt[kk + 1] & LOWER_MASK);
            mt[kk] = mt[kk + (M - N)] ^ (y >> 1) ^ mag01[y & 0x1UL];
        }

        y = (mt[N - 1] & UPPER_MASK) | (mt[0] & LOWER_MASK);
        mt[N - 1] = mt[M - 1] ^ (y >> 1) ^ mag01[y & 0x1UL];

        mti = 0;
    }
  
    y = mt[mti ++];

    /* Tempering */
    y ^= (y >> 11);
    y ^= (y << 7) & 0x9d2c5680UL;
    y ^= (y << 15) & 0xefc60000UL;
    y ^= (y >> 18);

    return y;
}

/* -------------------------------------------------------------------------- */

public unsigned long random_int32r(m_random *s)
{
    /** @brief generates a random number on [0,0xffffffff]-interval (thread safe) */

    unsigned long y;

    if (s->_mti >= N) { /* generate N words at one time */
        int kk;

        if (s->_mti == N + 1)   /* if random_init() has not been called, */
            random_init(5489UL); /* a default initial seed is used */

        for (kk = 0; kk < N - M; kk ++) {
            y = (s->_mt[kk] & UPPER_MASK) | (s->_mt[kk + 1] & LOWER_MASK);
            s->_mt[kk] = s->_mt[kk + M] ^ (y >> 1) ^ s->_mag01[y & 0x1UL];
        }

        for ( ; kk < N - 1; kk ++) {
            y = (s->_mt[kk] & UPPER_MASK) | (s->_mt[kk + 1] & LOWER_MASK);
            s->_mt[kk] = s->_mt[kk + (M - N)] ^ (y >> 1) ^ s->_mag01[y & 0x1UL];
        }

        y = (s->_mt[N - 1] & UPPER_MASK) | (s->_mt[0] & LOWER_MASK);
        s->_mt[N - 1] = s->_mt[M - 1] ^ (y >> 1) ^ s->_mag01[y & 0x1UL];

        s->_mti = 0;
    }
  
    y = s->_mt[s->_mti ++];

    /* Tempering */
    y ^= (y >> 11);
    y ^= (y << 7) & 0x9d2c5680UL;
    y ^= (y << 15) & 0xefc60000UL;
    y ^= (y >> 18);

    return y;
}

/* -------------------------------------------------------------------------- */

public inline long random_int31(void)
{
    /** @brief generates a random number on [0,0x7fffffff]-interval */

    return (long)(random_int32() >> 1);
}

/* -------------------------------------------------------------------------- */

public inline long random_int31r(m_random *state)
{
    /** @brief generates a random number on [0,0x7fffffff]-interval (thread safe) */

    return (long)(random_int32r(state) >> 1);
}

/* -------------------------------------------------------------------------- */
/* These real versions are due to Isaku Wada, 2002/01/09 added */
/* -------------------------------------------------------------------------- */

public inline double random_real1(void)
{
    /** @brief generates a random number on [0,1]-real-interval */

    return random_int32() * (1.0/4294967295.0);
    /* divided by 2^32-1 */
}

/* -------------------------------------------------------------------------- */

public inline double random_real1r(m_random *state)
{
    /** @brief generates a random number on [0,1]-real-interval (thread safe) */

    return random_int32r(state) * (1.0/4294967295.0);
    /* divided by 2^32-1 */
}

/* -------------------------------------------------------------------------- */

public inline double random_real2(void)
{
    /** @brief generates a random number on [0,1)-real-interval */

    return random_int32() * (1.0/4294967296.0);
    /* divided by 2^32 */
}

/* -------------------------------------------------------------------------- */

public inline double random_real2r(m_random *state)
{
    /** @brief generates a random number on [0,1)-real-interval (thread safe) */

    return random_int32r(state) * (1.0/4294967296.0);
    /* divided by 2^32 */
}

/* -------------------------------------------------------------------------- */

public inline double random_real3(void)
{
    /** @brief generates a random number on (0,1)-real-interval */

    return (((double) random_int32()) + 0.5) * (1.0/4294967296.0);
    /* divided by 2^32 */
}

/* -------------------------------------------------------------------------- */

public inline double random_real3r(m_random *state)
{
    /** @brief generates a random number on (0,1)-real-interval (thread safe) */

    return (((double) random_int32r(state)) + 0.5) * (1.0/4294967296.0);
    /* divided by 2^32 */
}

/* -------------------------------------------------------------------------- */

public inline double random_res53(void)
{
    /** @brief generates a random number on [0,1) with 53-bit resolution */

    unsigned long a = random_int32() >> 5, b = random_int32() >> 6;

    return (a * 67108864.0 + b) * (1.0/9007199254740992.0);
}

/* -------------------------------------------------------------------------- */

public inline double random_res53r(m_random *state)
{
    /**
     * @brief generates a random number on [0,1) with 53-bit resolution (thread safe)
     */

    unsigned long a = random_int32r(state) >> 5, b = random_int32r(state) >> 6;

    return (a * 67108864.0 + b) * (1.0/9007199254740992.0);
}

/* -------------------------------------------------------------------------- */
