#ifndef M_RANDOM_H

#define M_RANDOM_H

#include "m_core_def.h"

/** @defgroup random core::random */

typedef struct m_random {
    /* private */
    unsigned long *_mt;
    int _mti;
    unsigned long _mag01[2];
} m_random;

/**
 * @ingroup random
 * @struct m_random
 *
 * This structure stores the internal state of the Mersenne Twister random
 * number generator.
 *
 * Its content is private and should never be tempered with.
 *
 * @b private @ref _mt is the internal state vector
 * @b private @ref _mti is the internal state index
 * @b private @ref _mag01 = x * MATRIX_A  for x=0,1
 * The m_random structure is used to keep the state of the random numbers
 * generator for the thread safe API.
 *
 * You can either allocate it or get one using @ref random_initr()
 * It *must* be properly destroyed after use with @ref random_finir()
 *
 */

/* -------------------------------------------------------------------------- */

public void random_init(unsigned long s);

/**
 * @ingroup random
 * @fn void random_init(unsigned long s)
 * @param s the seed
 * @return void
 *
 * This function initializes the random number generator with the given seed.
 *
 */

/* -------------------------------------------------------------------------- */

public void random_arrayinit(const unsigned long init_key[],
                             unsigned long key_length       );

/**
 * @ingroup random
 * @fn void random_arrayinit(const unsigned long init_key[], unsigned long key_length)
 * @param init_key the key
 * @param key_length the length of the key
 * @return void
 *
 * This function initializes the random number generator with the given array.
 *
 */

/* -------------------------------------------------------------------------- */

public unsigned long random_int32(void);

/**
 * @ingroup random
 * @fn unsigned long random_int32(void)
 * @param void
 * @return an unsigned 32 bits integer
 *
 * This function generates a random number on [0,0xffffffff]-interval.
 *
 */

/* -------------------------------------------------------------------------- */

public inline long random_int31(void);

/**
 * @ingroup random
 * @fn long random_int31(void)
 * @param void
 * @return a signed, positive 32 bits integer
 *
 * This function generates a random number on [0,0x7fffffff]-interval.
 *
 */

/* -------------------------------------------------------------------------- */

public inline double random_real1(void);

/**
 * @ingroup random
 * @fn double random_real1(void)
 * @param void
 * @return a real
 *
 * This function generates a random number on [0,1]-real-interval.
 *
 */

/* -------------------------------------------------------------------------- */

public inline double random_real2(void);

/**
 * @ingroup random
 * @fn double random_real2(void)
 * @param void
 * @return a real
 *
 * This function generates a random number on [0,1)-real-interval.
 *
 */

/* -------------------------------------------------------------------------- */

public inline double random_real3(void);

/**
 * @ingroup random
 * @fn double random_real3(void)
 * @param void
 * @return a real
 *
 * This function generates a random number on (0,1)-real-interval.
 *
 */

/* -------------------------------------------------------------------------- */

public inline double random_res53(void);

/**
 * @ingroup random
 * @fn double random_res53(void)
 * @param void
 * @return a real
 *
 * This function generates a random number on [0,1) with 53-bit resolution.
 *
 */

/* -------------------------------------------------------------------------- */
/* Thread safe API */
/* -------------------------------------------------------------------------- */

public m_random *random_initr(unsigned long s);

/**
 * @ingroup random
 * @fn m_random *random_initr(unsigned long s)
 * @param s the seed
 * @return an opaque structure holding the random number generator state.
 *
 * This function initializes the random number generator with the given seed.
 * Thread safe version of @ref random_init().
 *
 * @see random_init()
 *
 */

/* -------------------------------------------------------------------------- */

public inline void random_finir(m_random *state);

/**
 * @ingroup random
 * @fn void random_finir(m_random *state)
 * @param state an opaque structure holding the random number generator state
 * @return void
 *
 * This function destroys a structure returned by @ref random_initr() or
 * @ref random_arrayinitr()
 *
 */

/* -------------------------------------------------------------------------- */

public m_random *random_arrayinitr(const unsigned long init_key[],
                                   unsigned long key_length       );

/**
 * @ingroup random
 * @fn m_random *random_arrayinitr(const unsigned long init_key[],
 *                                unsigned long key_length       )
 * @param init_key the key
 * @param key_length the length of the key
 * @return an opaque structure holding the random number generator state
 *
 * This function initializes the random number generator with the given array.
 * Thread safe version of @ref random_arrayinit().
 *
 * @see random_arrayinit()
 *
 */

/* -------------------------------------------------------------------------- */

public unsigned long random_int32r(m_random *state);

/**
 * @ingroup random
 * @fn unsigned long random_int32r(m_random *state)
 * @param state an opaque structure holding the random number generator state
 * @return an unsigned 32 bits integer
 *
 * This function generates a random number on [0,0xffffffff]-interval.
 * Thread safe version of @ref random_int32().
 *
 * @see random_int32()
 *
 */

/* -------------------------------------------------------------------------- */

public inline long random_int31r(m_random *state);

/**
 * @ingroup random
 * @fn long random_int31r(m_random *state)
 * @param state an opaque structure holding the random number generator state
 * @return a signed, positive 32 bits integer
 *
 * This function generates a random number on [0,0x7fffffff]-interval.
 * Thread safe version of @ref random_int31().
 *
 * @see random_int31()
 *
 */

/* -------------------------------------------------------------------------- */

public inline double random_real1r(m_random *state);

/**
 * @ingroup random
 * @fn double random_real1r(m_random *state)
 * @param state an opaque structure holding the random number generator state
 * @return a real
 *
 * This function generates a random number on [0,1]-real-interval.
 * Thread safe version of @ref random_real1().
 *
 * @see random_real1()
 *
 */

/* -------------------------------------------------------------------------- */

public inline double random_real2r(m_random *state);

/**
 * @ingroup random
 * @fn double random_real2r(m_random *state)
 * @param state an opaque structure holding the random number generator state
 * @return a real
 *
 * This function generates a random number on [0,1)-real-interval.
 * Thread safe version of @ref random_real2().
 *
 * @see random_real2()
 *
 */

/* -------------------------------------------------------------------------- */

public inline double random_real3r(m_random *state);

/**
 * @ingroup random
 * @fn double random_real3(void)
 * @param state an opaque structure holding the random number generator state
 * @return a real
 *
 * This function generates a random number on (0,1)-real-interval.
 * Thread safe version of @ref random_real3().
 *
 * @see random_real3()
 *
 */

/* -------------------------------------------------------------------------- */

public inline double random_res53r(m_random *state);

/**
 * @ingroup random
 * @fn double random_res53r(m_random *state)
 * @param state an opaque structure holding the random number generator state
 * @return a real
 *
 * This function generates a random number on [0,1) with 53-bit resolution.
 * Thread safe version of @ref random_res53().
 *
 * @see random_res53()
 *
 */

/* -------------------------------------------------------------------------- */

#endif
