#include "m_queue.h"

typedef struct _m_node {
    void *_data;
    struct _m_node *_next;
} _m_node;

/* -------------------------------------------------------------------------- */

private m_queue *queue_alloc(void)
{
    m_queue *ret = malloc(sizeof(*ret));

    if (! ret) { perror(ERR(queue_alloc, malloc)); return NULL; }

    if (pthread_mutex_init(& ret->_head_lock, NULL) == -1) {
        perror(ERR(queue_alloc, pthread_mutex_init)); return queue_free(ret);
    }

    if (pthread_mutex_init(& ret->_tail_lock, NULL) == -1) {
        perror(ERR(queue_alloc, pthread_mutex_init)); return queue_free(ret);
    }

    if ( (ret->_head = ret->_tail = malloc(sizeof(_m_node))) == NULL) {
        perror(ERR(queue_alloc, malloc)); return queue_free(ret);
    }

    ret->_head->_next = ret->_head->_data = NULL;

    return ret;
}

/* -------------------------------------------------------------------------- */

private m_queue *queue_free(m_queue *queue)
{
    if (queue) {
        pthread_mutex_destroy(& queue->_head_lock);
        pthread_mutex_destroy(& queue->_tail_lock);
        free(queue->_head);
    }

    free(queue);

    return NULL;
}

/* -------------------------------------------------------------------------- */

private void queue_free_nodes(m_queue *queue, void (*free_data)(void *))
{
    void *ptr = NULL;

    if (queue && free_data)
        while ( (ptr = queue_pop(queue)) != NULL) free_data(ptr);
}

/* -------------------------------------------------------------------------- */

private int queue_push(m_queue *queue, void *ptr)
{
    _m_node *node = NULL;

    if (! queue || ! ptr) return -1;

    if ( (node = malloc(sizeof(*node))) == NULL) {
        perror(ERR(queue_push, malloc)); return -1;
    }

    /* use the Michael & Scott two-lock concurrent queue algorithm */
    node->_data = ptr; node->_next = NULL;

    pthread_mutex_lock(& queue->_tail_lock);

        /* swing the tail to the new node */
        queue->_tail->_next = node;
        queue->_tail = node;

    pthread_mutex_unlock(& queue->_tail_lock);

    return 0;
}

/* -------------------------------------------------------------------------- */

private int queue_empty(m_queue *queue)
{
    int ret = 0;

    if (! queue) return 0;

    pthread_mutex_lock(& queue->_head_lock);

        ret = (queue->_head->_next == NULL) ? 0 : 1;

    pthread_mutex_unlock(& queue->_head_lock);

    return ret;
}

/* -------------------------------------------------------------------------- */

private void *queue_pop(m_queue *queue)
{
    _m_node *node = NULL, *next = NULL;
    void *ret = NULL;

    if (! queue) return NULL;

    /* use the Michael & Scott two-lock concurrent queue algorithm */
    pthread_mutex_lock(& queue->_head_lock);

        node = queue->_head; next = node->_next;

        /* pop the head and fetch the data */
        if (next) {
            ret = next->_data;
            free(node);
            queue->_head = next;
        }

    pthread_mutex_unlock(& queue->_head_lock);

    return ret;
}

/* -------------------------------------------------------------------------- */

private int queue_unpop(m_queue *queue, void *ptr)
{
    _m_node *node = NULL;

    if (! queue || ! ptr) return -1;

    if ( (node = malloc(sizeof(*node))) == NULL) {
        perror(ERR(queue_unpop, malloc)); return -1;
    }

    node->_data = NULL;

    pthread_mutex_lock(& queue->_head_lock);

        /* store the data in the head replace it with a dummy node */
        queue->_head->_data = ptr;
        node->_next = queue->_head;
        queue->_head = node;

    pthread_mutex_unlock(& queue->_head_lock);

    return 0;
}

/* -------------------------------------------------------------------------- */
