#ifndef M_QUEUE_H

#define M_QUEUE_H

#include "m_core_def.h"

/** @defgroup queue core::queue */

typedef struct m_queue {
    /* private */
    pthread_mutex_t _head_lock;
    pthread_mutex_t _tail_lock;

    struct _m_node *_head;
    struct _m_node *_tail;
} m_queue;

/**
 * @ingroup queue
 * @struct m_queue
 *
 * This structure holds the inner data of a two-lock concurrent queue.
 *
 * Its fields are private and should never be directly accessed.
 *
 * This said, here are their descriptions:
 *
 * @b private @ref _head_lock is the mutex protecting the head of the queue.
 * @b private @ref _tail_lock is the mutex protecting the tail of the queue.
 * @b private @ref _head is the head of the queue.
 * @b private @ref _head is the tail of the queue.
 *
 * The head and tail are of the private type _m_node, which is not exported.
 * This type may be modified at any time, so it should never be used outside
 * this module.
 *
 * However, here is the documentation of the _m_node structure:
 *
 * The type _m_node holds the data pushed onto the queue. It has currently
 * the following definition:
 *
 * typedef struct _m_node {
 *   void *_data;
 *   struct _m_node *_next;
 * } _m_node;
 *
 * @b private _data is used to carry a pointer to any data structure, or any
 * integer type which size is less than sizeof(int *).
 * @b private _next is used to link the nodes of the queue.
 *
 * Practically, the queue module has been designed to be extended for use
 * with your own data types. You can build a queue of a specific data type
 * simply by writing some wrappers to the generic functions provided.
 *
 */

/* -------------------------------------------------------------------------- */

private m_queue *queue_alloc(void);

/**
 * @ingroup queue
 * @fn m_queue *queue_alloc(void)
 * @param void
 * @return a pointer to a new m_queue, or NULL.
 *
 * This function allocates a new fully initialized queue.
 *
 */

/* -------------------------------------------------------------------------- */

private m_queue *queue_free(m_queue *queue);

/**
 * @ingroup queue
 * @fn m_queue *queue_free(m_queue *queue)
 * @param queue a pointer to a queue
 * @return always NULL
 *
 * This function will properly clean up and destroy a m_queue structure.
 * However, the nodes the queue may still be holding are not destroyed, since
 * queue_free has no mean to know how to properly handle the data inside the
 * nodes. To clean the nodes, please see @ref queue_free_nodes().
 *
 * Since this function does not empty the queue before destroying it, it must
 * only be used on empty queues (please see @ref queue_empty() ).
 *
 * This function always returns NULL, so it could be used to clean a pointer:
 * junk = queue_free(junk);
 *
 */

/* -------------------------------------------------------------------------- */

private void queue_free_nodes(m_queue *queue, void (*free_data)(void *));

/**
 * @ingroup queue
 * @fn void queue_free_nodes(m_queue *queue, void (*free_data)(void *))
 * @param queue a pointer to a queue
 * @param free_data a callback the destruction of the data inside the queue
 * @return void
 *
 * This function destroys all the nodes of a queue and clean their content
 * with the provided @b free_data callback.
 *
 */

/* -------------------------------------------------------------------------- */

private int queue_push(m_queue *queue, void *ptr);

/* -------------------------------------------------------------------------- */

private int queue_empty(m_queue *queue);

/* -------------------------------------------------------------------------- */

private void *queue_pop(m_queue *queue);

/* -------------------------------------------------------------------------- */

private int queue_unpop(m_queue *queue, void *ptr);

/* -------------------------------------------------------------------------- */

#endif
