#ifndef M_RANDOM_H

#define M_RANDOM_H

#include "m_core_def.h"

/*
------------------------------------------------------------------------------
rand.h: definitions for a random number generator
By Bob Jenkins, 1996, Public Domain
MODIFIED:
  960327: Creation (addition of randinit, really)
  970719: use context, not global variables, for internal state
  980324: renamed seed to flag
  980605: recommend RANDSIZL=4 for noncryptography.
  010626: note this is public domain
------------------------------------------------------------------------------
*/

/*
 * This algorithm is as fast (even 20% faster with gcc -O3) as the
 * Mersenne Twister, but it is considered cryptographically secure
 * and is natively reentrant :)
 *
 */

/* -------------------------------------------------------------------------- */
/* 0. Standard definitions */
/* -------------------------------------------------------------------------- */

typedef  unsigned long long  ub8;
#define UB8MAXVAL 0xffffffffffffffffLL
#define UB8BITS 64
typedef    signed long long  sb8;
#define SB8MAXVAL 0x7fffffffffffffffLL
typedef  unsigned long  int  ub4;   /* unsigned 4-byte quantities */
#define UB4MAXVAL 0xffffffff
typedef    signed long  int  sb4;
#define UB4BITS 32
#define SB4MAXVAL 0x7fffffff
typedef  unsigned short int  ub2;
#define UB2MAXVAL 0xffff
#define UB2BITS 16
typedef    signed short int  sb2;
#define SB2MAXVAL 0x7fff
typedef  unsigned       char ub1;
#define UB1MAXVAL 0xff
#define UB1BITS 8
typedef    signed       char sb1;   /* signed 1-byte quantities */
#define SB1MAXVAL 0x7f
typedef                 int  word;  /* fastest type available */

#define bis(target,mask)  ((target) |=  (mask))
#define bic(target,mask)  ((target) &= ~(mask))
#define bit(target,mask)  ((target) &   (mask))
#ifndef min
# define min(a,b) (((a)<(b)) ? (a) : (b))
#endif /* min */
#ifndef max
# define max(a,b) (((a)<(b)) ? (b) : (a))
#endif /* max */
#ifndef align
# define align(a) (((ub4)a+(sizeof(void *)-1))&(~(sizeof(void *)-1)))
#endif /* align */
#ifndef abs
# define abs(a)   (((a)>0) ? (a) : -(a))
#endif
#define TRUE  1
#define FALSE 0
#define SUCCESS 0  /* 1 on VAX */

/* -------------------------------------------------------------------------- */

/** @defgroup random core::random */

#define RANDSIZL   (8)  /* I recommend 8 for crypto, 4 for simulations */
#define RANDSIZ    (1<<RANDSIZL)

/* context of random number generator */
typedef struct m_random {
  ub4 _cnt;
  ub4 _rsl[RANDSIZ];
  ub4 _mem[RANDSIZ];
  ub4 _a;
  ub4 _b;
  ub4 _c;
} m_random;

/**
 * @ingroup random
 * @struct m_random
 *
 * This structure stores the internal state of the ISAAC random number
 * generator.
 *
 * Its content is private and should never be tampered with.
 *
 * @b private @ref randcnt is the number of generated numbers available
 * @b private @ref randrsl holds the results
 * @b private @ref randmem is the working memory
 * @b private @ref randa, @ref randb, @ref randc are working variables
 *
 * You get a new structure using random_init(), and you must destroy it using
 * random_fini().
 *
 */

/* -------------------------------------------------------------------------- */

public inline m_random *random_init(void);

/**
 * @ingroup random
 * @fn m_random *random_init(void)
 * @param void
 * @return an opaque structure holding the random number generator state
 *
 * This function initializes the random number generator with a default seed.
 *
 */

/* -------------------------------------------------------------------------- */

public m_random *random_arrayinit(const ub4 *seed, size_t len);

/**
 * @ingroup random
 * @fn m_random *random_arrayinit(const ub4 *seed, size_t len)
 * @param init_key the key
 * @param key_length the length of the key
 * @return an opaque structure holding the random number generator state
 *
 * This function initializes the random number generator with the given array.
 *
 */

/* -------------------------------------------------------------------------- */

public inline unsigned long random_uint32(m_random *ctx);

/**
 * @ingroup random
 * @fn unsigned long random_uint32(m_random *ctx)
 * @param ctx the random number generator state obtained from random_*init()
 * @return an unsigned 32 bits integer
 *
 * This function generates a random number on [0,0xffffffff]-interval.
 *
 * @warning If @ref ctx is NULL or invalid, the function will trigger a
 *          well deserved SIGSEGV. You have been warned ;)
 *
 */

/* -------------------------------------------------------------------------- */

public inline long random_int32(m_random *ctx);

/**
 * @ingroup random
 * @fn long random_int32(m_random *ctx)
 * @param ctx the random number generator state obtained from random_*init()
 * @return a signed, positive 32 bits integer
 *
 * This function generates a random number on [0,0x7fffffff]-interval.
 *
 * @warning If @ref ctx is NULL or invalid, the function will trigger a
 *          well deserved SIGSEGV. You have been warned ;)
 *
 */

/* -------------------------------------------------------------------------- */

public inline double random_real1(m_random *ctx);

/**
 * @ingroup random
 * @fn double random_real1(m_random *ctx)
 * @param ctx the random number generator state obtained from random_*init()
 * @return a real
 *
 * This function generates a random number on [0,1]-real-interval.
 *
 * @warning If @ref ctx is NULL or invalid, the function will trigger a
 *          well deserved SIGSEGV. You have been warned ;)
 *
 */

/* -------------------------------------------------------------------------- */

public inline double random_real2(m_random *ctx);

/**
 * @ingroup random
 * @fn double random_real2(m_random *ctx)
 * @param ctx the random number generator state obtained from random_*init()
 * @return a real
 *
 * This function generates a random number on [0,1)-real-interval.
 *
 * @warning If @ref ctx is NULL or invalid, the function will trigger a
 *          well deserved SIGSEGV. You have been warned ;)
 *
 */

/* -------------------------------------------------------------------------- */

public inline double random_real3(m_random *ctx);

/**
 * @ingroup random
 * @fn double random_real3(m_random *ctx)
 * @param ctx the random number generator state obtained from random_*init()
 * @return a real
 *
 * This function generates a random number on (0,1)-real-interval.
 *
 * @warning If @ref ctx is NULL or invalid, the function will trigger a
 *          well deserved SIGSEGV. You have been warned ;)
 *
 */

/* -------------------------------------------------------------------------- */

public inline double random_res53(m_random *ctx);

/**
 * @ingroup random
 * @fn double random_res53(m_random *ctx)
 * @param ctx the random number generator state obtained from random_*init()
 * @return a real
 *
 * This function generates a random number on [0,1) with 53-bit resolution.
 *
 * @warning If @ref ctx is NULL or invalid, the function will trigger a
 *          well deserved SIGSEGV. You have been warned ;)
 *
 */

/* -------------------------------------------------------------------------- */

#endif
