#include "m_socket.h"

/* macro for printing getaddrinfo error messages */
#define _gai_perror(s, i) (fprintf(stderr, "%s: %s\n", (s), gai_strerror((i))))

#ifdef _ENABLE_SSL
/* macro for printing SSL error messages */
#define sslerror(s) \
(fprintf(stderr, "%s: %s\n", (s), ERR_reason_error_string(ERR_get_error())))
#endif

static pthread_mutex_t _id_lock = PTHREAD_MUTEX_INITIALIZER;
static uint16_t _id = 1;
static unsigned int _free_count = 0;
static uint16_t _free_id[32768];

/**
 * @var _socket
 *
 * This private array holds all the allocated sockets. It is used for direct
 * access using dedicated functions, but it MUST NEVER be used without proper
 * locking at the array *and* socket levels.
 *
 * Never forget that a server thread may be using the socket you are trying to
 * fiddle with !
 *
 */

static pthread_rwlock_t _socket_lock = PTHREAD_RWLOCK_INITIALIZER;
static m_socket *_socket[32768];

#ifdef _ENABLE_SSL
/* OpenSSL requires some static initialization */
static int _openssl_initialized = 0;
static SSL_CTX *_ssl_ctx;
static int _sc = 1;

/* SSL helpers */
static void _socket_ssl_init(void);
static void _socket_ssl_fini(void);
static m_socket *_socket_ssl_open(m_socket *s);
static ssize_t _socket_ssl_write(m_socket *s, const char *data, size_t len);
static ssize_t _socket_ssl_read(m_socket *s, char *out, size_t len);
static int password_cb(char *buf, int len, int rwflag,void *userdata);
#endif

/* -------------------------------------------------------------------------- */

static struct addrinfo *_socket_addr(int *flags, const char *ip, const char *pt)
{
    /* this helper builds a socket address using flags, an ip and a port */

    struct addrinfo hint, *res = NULL;
    int err = 0;

    if (! flags) return NULL;

    /* if no port was given, set it to 0 and ensure we will not listen */
    if (! pt) { pt = "0"; *flags &= ~SOCKET_SRV; }

    /* fill the hint structure for getaddrinfo() */
    memset(& hint, 0, sizeof(hint));
    hint.ai_family   = (*flags & SOCKET_IP6) ? PF_INET6 : PF_INET;
    hint.ai_flags    = AI_NUMERICHOST | AI_PASSIVE;
    hint.ai_socktype = (*flags & SOCKET_UDP) ? (SOCK_DGRAM) : (SOCK_STREAM);
    hint.ai_protocol = (*flags & SOCKET_UDP) ? (IPPROTO_UDP) : (IPPROTO_TCP);

    /* if no ip was given and a client is requested, remove AI_PASSIVE */
    if (! ip && (~*flags & SOCKET_SRV)) hint.ai_flags &= ~AI_PASSIVE;

    /* use getaddrinfo() to build the socket address */
    if ( (err = getaddrinfo(ip, pt, & hint, & res)) != 0)
        _gai_perror(ERR(_socket_addr, getaddrinfo), err);

    return res;
}

/* -------------------------------------------------------------------------- */
/* RESTRICTED SSL initialization/cleanup functions */
/* -------------------------------------------------------------------------- */

#ifdef _ENABLE_SSL
static int password_cb(char *buf, int len, unused int rwflag,
                       unused void *userdata                 )
{
    if (len < (int) strlen("password") + 1) return 0;

    memcpy(buf, "password", strlen("password") + 1);

    return(strlen("password"));
}

/* -------------------------------------------------------------------------- */

static void _socket_ssl_init(void)
{
    SSL_METHOD *meth = NULL;

    if (_openssl_initialized) return;

    SSL_library_init();
    SSL_load_error_strings();
    OpenSSL_add_all_algorithms();

    if (! (meth = SSLv23_method()) ) {
        sslerror(ERR(_socket_ssl_init, SSLv23_method));
        exit(EXIT_FAILURE);
    }

    if (! (_ssl_ctx = SSL_CTX_new(meth)) ) {
        sslerror(ERR(_socket_ssl_init, SSL_CTX_new));
        exit(EXIT_FAILURE);
    }

    if (! SSL_CTX_use_certificate_chain_file(_ssl_ctx, "../ssl/client.pem")) {
        sslerror(ERR(_socket_ssl_init, SSL_CTX_use_certificate_chain_file));
        exit(EXIT_FAILURE);
    }

    SSL_CTX_set_default_passwd_cb(_ssl_ctx, password_cb);

    if (! SSL_CTX_use_PrivateKey_file(_ssl_ctx, "../ssl/client.pem", SSL_FILETYPE_PEM)) {
        sslerror(ERR(_socket_ssl_init, SSL_CTX_use_PrivateKey_file));
        exit(EXIT_FAILURE);
    }

    if (! SSL_CTX_load_verify_locations(_ssl_ctx, "../ssl/root.pem", 0)) {
        sslerror(ERR(_socket_ssl_init, SSL_CTX_load_verify_locations));
        exit(EXIT_FAILURE);
    }

    #if (OPENSSL_VERSION_NUMBER < 0x00905100L)
    SSL_CTX_set_verify_depth(_ssl_ctx, 1);
    #endif 

    SSL_CTX_set_session_id_context(_ssl_ctx, (void *) & _sc, sizeof(_sc));

    atexit(_socket_ssl_fini);

    _openssl_initialized = 1;
}

/* -------------------------------------------------------------------------- */

static void _socket_ssl_fini(void)
{
    SSL_CTX_free(_ssl_ctx);
    ERR_free_strings();
}

/* -------------------------------------------------------------------------- */

static m_socket *_socket_ssl_open(m_socket *s)
{
    BIO *bio = NULL;
    if (! s || ~s->_flags & SOCKET_SSL) return NULL;

    if (! (s->_ssl = SSL_new(_ssl_ctx)) ) {
        sslerror(ERR(_socket_ssl_open, SSL_new));
        return socket_close(s);
    }
    
    if (! (bio = BIO_new_socket(s->_fd, BIO_NOCLOSE)) ) {
        sslerror(ERR(_socket_ssl_open, BIO_new_socket));
        return socket_close(s);
    }
    
    SSL_set_bio(s->_ssl, bio, bio);

    if (s->_flags & SOCKET_SRV)
        SSL_set_accept_state(s->_ssl);
    else
        SSL_set_connect_state(s->_ssl);

    return s;
}

/* -------------------------------------------------------------------------- */

static int _socket_ssl_connect(m_socket *s)
{
    int ret = 0;

    if (! s || ~s->_flags & SOCKET_SSL) return SOCKET_EPARAM;

    /* try to connect */
    ret = SSL_connect(s->_ssl);

    switch (SSL_get_error(s->_ssl, ret)) {
        /* all went fine */
        case SSL_ERROR_NONE:
            s->_flags &= ~(SOCKET_BRD | SOCKET_BWR | SOCKET_CON);
            return 0;
        /* recoverable errors */
        case SSL_ERROR_WANT_READ:
            s->_flags |= (SOCKET_BRD | SOCKET_CON);
            return SOCKET_EAGAIN;
        case SSL_ERROR_WANT_WRITE:
            s->_flags |= (SOCKET_BWR | SOCKET_CON);
            return SOCKET_EAGAIN;
        /* link closed */
        case SSL_ERROR_ZERO_RETURN:
            return SOCKET_ECLOSE;
        default:
            sslerror(ERR(_socket_ssl_connect, SSL_connect));
            return SOCKET_EFATAL;
    }

    /* NOTREACHED */
    return 0;
}

/* -------------------------------------------------------------------------- */

static int _socket_ssl_accept(m_socket *s)
{
    int ret = 0;

    if (! s || ~s->_flags & SOCKET_SSL) return SOCKET_EPARAM;

    /* try to accept the connection */
    ret = SSL_accept(s->_ssl);

    switch (SSL_get_error(s->_ssl, ret)) {
        /* all went fine */
        case SSL_ERROR_NONE:
            s->_flags &= ~(SOCKET_BRD | SOCKET_BWR | SOCKET_ACC);
            return 0;
        /* recoverable errors */
        case SSL_ERROR_WANT_READ:
            s->_flags |= (SOCKET_BRD | SOCKET_ACC);
            return SOCKET_EAGAIN;
        case SSL_ERROR_WANT_WRITE:
            s->_flags |= (SOCKET_BWR | SOCKET_ACC);
            return SOCKET_EAGAIN;
        /* link closed */
        case SSL_ERROR_ZERO_RETURN:
            return SOCKET_ECLOSE;
        default:
            sslerror(ERR(_socket_ssl_accept, SSL_accept));
            return SOCKET_EFATAL;
    }

    /* NOTREACHED */
    return 0;
}

/* -------------------------------------------------------------------------- */

static ssize_t _socket_ssl_write(m_socket *s, const char *data, size_t len)
{
    ssize_t ret = 0;

    /* check if there is a connect() or accept() in progress */
    if (s->_flags & SOCKET_CON) {
        if ( (ret = _socket_ssl_connect(s)) != 0)
            return ret;
    } else if (s->_flags & SOCKET_ACC) {
        if ( (ret = _socket_ssl_accept(s)) != 0)
            return ret;
    }

    ret = SSL_write(s->_ssl, data, len);

    switch (SSL_get_error(s->_ssl, ret)) {
        /* all went fine */
        case SSL_ERROR_NONE:
            s->_flags &= ~(SOCKET_BRD | SOCKET_BWR);
            return ret;
        /* recoverable errors */
        case SSL_ERROR_WANT_READ:
             s->_flags |= SOCKET_BRD;
            return SOCKET_EAGAIN;
        case SSL_ERROR_WANT_WRITE:
            s->_flags |= SOCKET_BWR;
            return SOCKET_EAGAIN;
        /* link closed */
        case SSL_ERROR_ZERO_RETURN:
            return SOCKET_ECLOSE;
        default:
            sslerror(ERR(_socket_ssl_write, SSL_write));
            return SOCKET_EFATAL;
    }

    /* NOTREACHED */
    return 0;
}

/* -------------------------------------------------------------------------- */

static ssize_t _socket_ssl_read(m_socket *s, char *out, size_t len)
{
    ssize_t ret = 0;

    /* check if there is a connect() or accept() in progress */
    if (s->_flags & SOCKET_CON) {
        if ( (ret = _socket_ssl_connect(s)) != 0)
            return ret;
    } else if (s->_flags & SOCKET_ACC) {
        if ( (ret = _socket_ssl_accept(s)) != 0)
            return ret;
    }

    ret = SSL_read(s->_ssl, out, len);

    switch (SSL_get_error(s->_ssl, ret)) {
        /* all went fine */
        case SSL_ERROR_NONE:
            s->_flags &= ~(SOCKET_BRD | SOCKET_BWR);
            return ret;
        /* recoverable errors */
        case SSL_ERROR_WANT_READ:
            s->_flags |= SOCKET_BRD;
            return SOCKET_EAGAIN;
        case SSL_ERROR_WANT_WRITE:
            s->_flags |= SOCKET_BWR;
            return SOCKET_EAGAIN;
        /* link closed */
        case SSL_ERROR_ZERO_RETURN:
            return SOCKET_ECLOSE;
        default:
            sslerror(ERR(_socket_ssl_read, SSL_read));
            return SOCKET_EFATAL;
    }

    /* NOTREACHED */
    return 0;
}

#endif

/* -------------------------------------------------------------------------- */
/* RESTRICTED, internal socket registration routines */
/* -------------------------------------------------------------------------- */

static int _socket_reg(m_socket *s, int type)
{
    uint16_t sockid = 0;

    /* check if there is a free socket slot */
    pthread_mutex_lock(& _id_lock);

    if (_free_count) {
        printf("_socket_reg(): using freelist for id\n");
        /* toggle the sign bit to help preventing race conditions */
        sockid = _free_id[-- _free_count] ^ 0x8000;
    } else if (_id < 32786) {
        printf("_socket_reg(): using identifier counter\n");
        /* allocate a new id */
        sockid = _id ++;
    }

    if (! sockid) {
        fprintf(stderr, "_socket_reg(): too many sockets.\n");
        return -1;
    }

    printf("_socket_reg(): using id=%i\n", sockid);

    pthread_mutex_unlock(& _id_lock);

    pthread_rwlock_wrlock(& _socket_lock);

    _socket[sockid & 0x7FFF] = s;

    pthread_rwlock_unlock(& _socket_lock);

    /* all went fine, record id in the upper 16 bits of the flags field */
    s->_flags = ( (sockid << 16) & 0xFFFF0000) | (type & 0xFFFF);

    printf("_socket_reg(): recorded id=0x%x type=0x%x\n",
           SOCKET_ID(s),
           (s->_flags & 0xFFFF)
          );

    return 0;
}

/* -------------------------------------------------------------------------- */

static m_socket *_socket_dereg(m_socket *s)
{
    m_socket *sock = NULL;

    if (! s) return NULL;

    /* first, remove the socket from the array to prevent further accesses */
    pthread_rwlock_wrlock(& _socket_lock);

    /* ignore the (bogus) sign bit */
    if (_socket[(s->_flags >> 16) & 0x7FFF] == s) {
        /* they match, fine */
        _socket[(s->_flags >> 16) & 0x7FFF] = NULL; sock = s;
    }

    pthread_rwlock_unlock(& _socket_lock);

    if (! sock) {
        /* the sockets do not match... something creepy is going on */
        fprintf(stderr, "_socket_dereg(): sockets do not match !\n");
        return NULL;
    }

    /* release the identifier */
    pthread_mutex_lock(& _id_lock);

    _free_id[_free_count ++] = SOCKET_ID(sock);

    pthread_mutex_unlock(& _id_lock);

    return sock;
}

/* -------------------------------------------------------------------------- */
/* Private, regular socket locking primitives */
/* -------------------------------------------------------------------------- */

private int socket_lock(m_socket *s)
{
    int ret = 0;

    if (! s) return -1;

    pthread_mutex_lock(s->_lock);

    /* sleep until someone release the lock */
    while (s->_lockstate == 0) pthread_cond_wait(s->_cond, s->_lock);

    /* we got woken up, check if the lockstate is valid */
    (s->_lockstate == -1) ? ret = -1 : s->_lockstate --;

    pthread_mutex_unlock(s->_lock);

    return ret;
}

/* -------------------------------------------------------------------------- */

private void socket_unlock(m_socket *s)
{
    if (! s) return;

    /* increment the lockstate if it is valid and signal the change */

    pthread_mutex_lock(s->_lock);

    if (s->_lockstate >= 0) s->_lockstate ++;

    pthread_mutex_unlock(s->_lock);

    pthread_cond_signal(s->_cond);
}

/* -------------------------------------------------------------------------- */
/* RESTRICTED socket lock breaking primitive */
/* -------------------------------------------------------------------------- */

static void _socket_break_lock(m_socket *s)
{
    pthread_mutex_lock(s->_lock);

    s->_lockstate = -1;

    pthread_mutex_unlock(s->_lock);

    pthread_cond_broadcast(s->_cond);
}

/* -------------------------------------------------------------------------- */
/* Private socket allocation code */
/* -------------------------------------------------------------------------- */

public m_socket *socket_open(const char *ip, const char *port, int type)
{
    m_socket *new = NULL;
    struct addrinfo *info = NULL;
    SOCKET sockfd = 0;
    char name[BUFSIZ];
    int nbio = 1;

    /* try to get socket informations */
    if (! (info = _socket_addr(& type, ip, port)) ) return NULL;

    /* get the socket descriptor */
    sockfd = socket(info->ai_family, info->ai_socktype, info->ai_protocol);
    if (sockfd == INVALID_SOCKET) {
        /* XXX don't forget to use serror() for socket functions ! */
        serror(ERR(socket_open, socket)); goto _err_sock;
    }

    printf("socket_open(): got fd=%i\n", sockfd);

    if (~type & SOCKET_BIO) {
        /* use non-blocking i/o (default) */
        if (ioctl(sockfd, FIONBIO, & nbio) == -1)
            serror(ERR(socket_open, ioctl));
        else
            printf("socket_open(): FIONBIO successfull\n");
    }

    #ifdef _ENABLE_SSL
    if (type & SOCKET_SSL) {
        if (! _openssl_initialized) _socket_ssl_init();
    }
    #endif

    /* allocate the socket structure */
    if (! (new = malloc(sizeof(*new))) ) {
        perror(ERR(socket_open, malloc)); goto _err_alloc;
    }

    /* all went fine for now, try to register the socket */
    if (_socket_reg(new, type) == -1) goto _err_buff;

    /* create the socket name */
    snprintf(name, sizeof(name), SOCKET_NAMFMT, SOCKET_ID(new));
    printf("socket_open(): buffer name = %s\n", name);

    /* allocate its buffer as shm to be able to share it with plugins */
    if (! (new->_buf = shm_alloc(name, SOCKET_BUFFER)) ) {
        perror(ERR(socket_open, shm_alloc)); goto _err_buff;
    }
    printf("socket_open(): buffer allocation successfull\n");

    /* initialize the inner semaphore */
    if (! (new->_lock = malloc(sizeof(*new->_lock))) ) {
        perror(ERR(socket_open, malloc)); goto _err_shm;
    }

    if (! (new->_cond = malloc(sizeof(*new->_cond))) ) {
        perror(ERR(socket_open, malloc)); goto _err_lock;
    }

    if (pthread_mutex_init(new->_lock, NULL) == -1) {
        perror(ERR(socket_open, pthread_mutex_init)); goto _err_mx;
    }

    if (pthread_cond_init(new->_cond, NULL) == -1) {
        perror(ERR(socket_open, pthread_cond_init)); goto _err_cn;
    }

    printf("socket_open(): semaphore allocation successfull\n");

    /* all went fine, initialize and register the struct */
    new->_lockstate = 1;
    new->_off = new->_len = 0;
    new->_fd = sockfd;
    new->info = info;

    printf("socket_open(): fd=%i\n", new->_fd);

    #ifdef _ENABLE_SSL
    if (type & SOCKET_SSL) return _socket_ssl_open(new);
    #endif

    return new;
_err_cn:
    pthread_mutex_destroy(new->_lock);
_err_mx:
    free(new->_cond);
_err_lock:
    free(new->_lock);
_err_shm:
    shm_free(name, new->_buf, SOCKET_BUFFER);
_err_buff:
    free(new);
_err_sock:
    close(sockfd);
_err_alloc:
    freeaddrinfo(info);
    return NULL;
}

/* -------------------------------------------------------------------------- */

public int socket_connect(m_socket *s)
{
    int ret = 0;

    /* the socket must be a client */
    if (! s || s->_flags & SOCKET_SRV) return -1;

    /* non blocking i/o */
    if (~s->_flags & SOCKET_BIO) {
        if (socket_lock(s) == 0) {
            ret = connect(s->_fd, s->info->ai_addr, s->info->ai_addrlen);

            if (ret == 0) {
                /* success ! */
                s->_flags &= ~(SOCKET_BRD | SOCKET_BWR | SOCKET_CON);
            } else if (ret == -1) {
                switch (ERRNO) {
                    case EINTR:
                    case EALREADY:
                    case EINPROGRESS:
                        goto _erragain;
                    default:
                        goto _errconnect;
                }
            }

            #ifdef _ENABLE_SSL
            if (s->_flags & SOCKET_SSL && (ret = _socket_ssl_connect(s)) != 0) {
                socket_unlock(s);
                return ret;
            }
            #endif

            socket_unlock(s);

        } else goto _errbroken;
                    
    } else {
        /* blocking i/o */
        if (socket_lock(s) == 0) {
            ret = connect(s->_fd, s->info->ai_addr, s->info->ai_addrlen);
            socket_unlock(s);
        } else goto _errbroken;
    }

    return 0;

    _erragain:
        socket_unlock(s);
        s->_flags |= (SOCKET_BRD | SOCKET_BWR | SOCKET_CON);
        return SOCKET_EAGAIN;
    _errbroken:
        fprintf(stderr, "socket_connect(): broken socket.\n");
        return SOCKET_EFATAL;
    _errconnect:
        socket_unlock(s);
        return SOCKET_EFATAL;
}

/* -------------------------------------------------------------------------- */

public int socket_listen(m_socket *s)
{
    int r = 1;
    struct sockaddr_in *a = NULL;

    /* the socket must be a server */
    if (! s || ~s->_flags & SOCKET_SRV) return -1;

    /* listen() does only make sense with SOCK_STREAM sockets,
       so make this call a no-op for UDP sockets */
    if (s->_flags & SOCKET_UDP) return 0;

    if (socket_lock(s) == -1) {
        fprintf(stderr, "socket_listen(): broken socket.\n");
        return -1;
    }

    /* bind the socket if necessary */
    if (~s->_flags & SOCKET_BND) {
        /* use SO_REUSEADDR to avoid binding problems */
        r = setsockopt(s->_fd, SOL_SOCKET, SO_REUSEADDR, & r, sizeof(r));
        if (r == -1) {
            serror(ERR(_socket_bind, setsockopt));
            return -1;
        }
    
        /* check if we need privileges to bind the socket */
        a = (struct sockaddr_in *) s->info->ai_addr;
        if (ntohs(a->sin_port) <= 1024) {
            fprintf(stderr,
            "socket_listen(): privileged binding not implemented yet.\n");
            return -1;
        } else {
            if (bind(s->_fd, s->info->ai_addr, s->info->ai_addrlen) == -1) {
                serror(ERR(_socket_bind, bind));
                return -1;
            }
        }
    
        /* all went fine, mark the socket as bound */
        s->_flags |= SOCKET_BND;
    }

    /* we may now call listen() */
    if (listen(s->_fd, SOMAXCONN) == -1) {
        serror(ERR(socket_listen, listen));
        return -1;
    }

    socket_unlock(s);

    return 0;
}

/* -------------------------------------------------------------------------- */

private m_socket *socket_accept(m_socket *s)
{
    m_socket *new = NULL;
    struct sockaddr *remote = NULL;
    size_t rlen = 0;
    int ret = 0;
    char name[BUFSIZ];

    if (socket_lock(s) == -1) {
        fprintf(stderr, "socket_connect(): broken socket.\n");
        return NULL;
    }

    /* allocate the remote address buffer */
    if (! (remote = malloc(sizeof(*remote))) ) {
        perror(ERR(socket_accept, malloc));
        goto _erralloc;
    }

    /* try to accept a connection */
    if ( (ret = accept(s->_fd, remote, & rlen)) == INVALID_SOCKET) {
        perror(ERR(socket_accept, accept));
        goto _erraccpt;
    }

    /* nice, we got one */
    if (! (new = malloc(sizeof(*new))) ) {
        perror(ERR(socket_accept, malloc));
        goto _erraccpt;
    }

    /* try to register the socket */
    if (_socket_reg(new, s->_flags) == -1) goto _errmsock;

    /* create the socket name */
    snprintf(name, sizeof(name), SOCKET_NAMFMT, SOCKET_ID(new));
    printf("socket_accept(): buffer name = %s\n", name);

    /* allocate the buffer as shm to be able to share it with plugins */
    if (! (new->_buf = shm_alloc(name, SOCKET_BUFFER)) ) {
        perror(ERR(socket_open, shm_alloc)); goto _errmsock;
    }
    printf("socket_accept(): buffer allocation successfull\n");

    /* initialize the inner semaphore */
    if (! (new->_lock = malloc(sizeof(*new->_lock))) ) {
        perror(ERR(socket_open, malloc)); goto _err_shm;
    }

    if (! (new->_cond = malloc(sizeof(*new->_cond))) ) {
        perror(ERR(socket_open, malloc)); goto _err_lock;
    }

    if (pthread_mutex_init(new->_lock, NULL) == -1) {
        perror(ERR(socket_open, pthread_mutex_init)); goto _err_mx;
    }

    if (pthread_cond_init(new->_cond, NULL) == -1) {
        perror(ERR(socket_open, pthread_cond_init)); goto _err_cn;
    }

    printf("socket_accept(): semaphore allocation successfull\n");

    /* all went fine, initialize and register the struct */
    new->_lockstate = 1;
    new->_off = new->_len = 0;

    if (! (new->info = malloc(sizeof(*new->info))) ) {
        perror(ERR(socket_accept, malloc));
        goto _err_cn;
    }

    /* disable the server and bound flags inherited from the parent */
    new->_flags &= ~(SOCKET_SRV | SOCKET_BND);
    new->_fd = ret;
    new->info->ai_addr = remote;
    new->info->ai_addrlen = rlen;

    #ifdef _ENABLE_SSL
    if (new->_flags & SOCKET_SSL && (new = _socket_ssl_open(new)) ) {
        /* try to accept the connection */
        switch (_socket_ssl_accept(new)) {
            /* all went fine */
            case 0:
            case SOCKET_EAGAIN:
                return new;
            default:
                return socket_close(new);
        }
    }
    #endif

    socket_unlock(s);

    return new;

_err_cn:
    pthread_mutex_destroy(new->_lock);
_err_mx:
    free(new->_cond);
_err_lock:
    free(new->_lock);
_err_shm:
    shm_free(name, new->_buf, SOCKET_BUFFER);
_errmsock:
    free(new);
_erraccpt:
    free(remote);
_erralloc:
    socket_unlock(s);
    return NULL;
}

/* -------------------------------------------------------------------------- */

private ssize_t socket_write_lowlevel(m_socket *s, const char *data, size_t len)
{
    ssize_t ret = 0;

    if (! s || ! data || ! len) return SOCKET_EPARAM;

    #ifdef _ENABLE_SSL
    if (s->_flags & SOCKET_SSL) return _socket_ssl_write(s, data, len);
    #endif

    /* check for a pending connection */
    if (s->_flags & SOCKET_CON && ( (ret = socket_connect(s)) != 0) )
        return ret;

    ret = sendto(
                 s->_fd,
                 data,
                 len,
                 0x0,
                 s->info->ai_addr,
                 s->info->ai_addrlen
                );

    if (ret == -1) {
        switch(ERRNO) {
            /* recuperable errors */
            case EINTR:
            case EAGAIN:
            case ESPIPE:
                return SOCKET_EAGAIN;
            /* fatal errors */
            default:
                return SOCKET_EFATAL;
        }
    } else if (ret == 0) {
        return SOCKET_ECLOSE;
    }

    return ret;
}

/* -------------------------------------------------------------------------- */

private ssize_t socket_read_lowlevel(m_socket *s, char *out, size_t len)
{
    ssize_t ret = 0;

    if (! s || ! out || ! len) return SOCKET_EPARAM;

    #ifdef _ENABLE_SSL
    if (s->_flags & SOCKET_SSL) return _socket_ssl_read(s, out, len);
    #endif

    /* check for a pending connection */
    if (s->_flags & SOCKET_CON && ( (ret = socket_connect(s)) != 0) )
        return ret;

    ret = recvfrom(
                   s->_fd,
                   out,
                   len,
                   0x0,
                   s->info->ai_addr,
                   & s->info->ai_addrlen
                  );

    if (ret == -1) {
        switch(ERRNO) {
            /* recuperable errors */
            case EINTR:
            case EAGAIN:
            case ESPIPE:
                return SOCKET_EAGAIN;
            /* fatal errors */
            default:
                return SOCKET_EFATAL;
        }
    } else if (ret == 0) {
        return SOCKET_ECLOSE;
    }

    return ret;
}

/* -------------------------------------------------------------------------- */

public m_socket *socket_close(m_socket *s)
{
    m_socket *sock = NULL;
    char name[BUFSIZ];

    if (! (sock = _socket_dereg(s)) ) return NULL;

    /* we found the corresponding socket, acquire it first */
    if (socket_lock(sock) == -1) {
        /* the socket is already broken, smells fishy... */
        fprintf(stderr, "socket_close(): socket already broken !\n");
        return NULL;
    }

    /* now that we own the socket, free the threads waiting after it */
    _socket_break_lock(sock);

    /* fine, now the socket is entirely ours - destroy it */
    snprintf(name, sizeof(name), SOCKET_NAMFMT, SOCKET_ID(sock));
    shm_free(name, sock->_buf, SOCKET_BUFFER);
    freeaddrinfo(sock->info);

    pthread_mutex_destroy(sock->_lock);
    pthread_cond_destroy(sock->_cond);
    free(sock->_lock); free(sock->_cond);

    #ifdef _ENABLE_SSL
    if (sock->_flags & SOCKET_SSL) {
        SSL_shutdown(sock->_ssl);
        SSL_free(sock->_ssl);
    }
    #endif

    closesocket(sock->_fd);

    /* sir, target successfully destroyed, sir ! \o> */
    free(sock);

    return NULL;
}

/* -------------------------------------------------------------------------- */
