#ifndef M_SOCKET

#define M_SOCKET

#include "m_core_def.h"
#include "m_string.h"
#include "m_queue.h"

/** @defgroup socket core::socket */

/*typedef m_work_queue m_queue;
typedef m_sock_queue m_queue;*/

/* socket type flags (< 0x100) */
#define SOCKET_UDP 0x01     /* UDP socket */
#define SOCKET_SRV 0x02     /* server socket */
#define SOCKET_BIO 0x04     /* use blocking I/O */

/* socket status flags (> 0xFF) */
#define SOCKET_BND 0x100    /* the socket is bound */
#define SOCKET_BRD 0x200    /* the socket blocks on reading */
#define SOCKET_BWR 0x300    /* the socket blocks on writing */
#define SOCKET_ACC 0x400    /* the socket is being accepted */
#define SOCKET_CON 0x800    /* the socket is being connected */

/* enable SOCKET_SSL only if _ENABLE_SSL is set */
#ifdef _ENABLE_SSL
    #define SOCKET_SSL 0x10
#else
    #define SOCKET_SSL 0x0
#endif

/* disable the SOCKET_IP6 flag if IPv6 is not supported */
#ifdef PF_INET6
    #define SOCKET_IP6 0x20
#else
    #define SOCKET_IP6 0x0
#endif

#define SOCKET_BUFFER 65536
#define SOCKET_NAMFMT "mammouth_socket_%i"

#define SOCKET_EPARAM -1
#define SOCKET_EAGAIN -2
#define SOCKET_ECLOSE -3
#define SOCKET_EFATAL -4

/* the upper 16 bits of the flag holds the socket identifier */
#define SOCKET_ID(s) ( ((s)->_flags >> 16) & 0xFFFF)

typedef struct m_socket {
    /* private, socket flags and descriptor */
    uint32_t _flags;
    SOCKET _fd;

    /* private, buffer pointer and state */
    unsigned char *_buf;
    /* good enough to cover a 65536 bytes wide buffer */
    uint16_t _off;
    uint16_t _len;

    /* private, locking */
    pthread_mutex_t *_lock;
    int _lockstate;
    pthread_cond_t *_cond;

    /* public */
    struct addrinfo *info;

    #ifdef _ENABLE_SSL
    /* private, ssl */
    SSL *_ssl;
    #endif
} m_socket;

typedef struct m_task {
    /* private, inline data */
    m_string *_header;
    m_string *_footer;

    /* private, file transfer fd, offset and length */
    int _fd;
    size_t _off;
    size_t _len;

    /* private, file transfer encoding or specific handling callback */
    m_string *(*_xfer)(int fd, size_t off, size_t len);
} m_task;

/* -------------------------------------------------------------------------- */

public m_socket *socket_open(const char *ip, const char *port, int type);

/**
 * @ingroup socket
 * @fn m_socket *socket_open(const char *ip, const char *port, int type)
 * @param ip the host ip address; remote for a client, local for a server
 * @param port the port to connect or listen to
 * @param type the socket type (TCP/UDP/SSL, client/server...)
 * @return a new socket or NULL if something went wrong
 *
 * This functions creates a new socket and registers it. You may call
 * socket_connect() or socket_listen() on it in order to use it.
 *
 * You may create different types of sockets:
 * - SOCKET_UDP or SOCKET_TCP according to the underlying transfer protocol
 *   you wish to use
 * - SOCKET_IP6 if you want to use IPv6
 * - SOCKET_SSL if you want to use secure sockets
 * - SOCKET_SRV to create a listener socket
 * - SOCKET_BIO to use blocking i/o
 * SOCKET_IP6 and SOCKET_SSL may be however disabled at compilation time.
 *
 * A socket opened with this function MUST be destroyed with socket_close().
 *
 * @see socket_close(), socket_connect(), socket_listen()
 *
 */

/* -------------------------------------------------------------------------- */

private int socket_lock(m_socket *s);

/**
 * @ingroup socket
 * @fn int socket_lock(m_socket *s)
 * @param s the socket to lock
 * @return 0 if the socket was successfully locked, -1 otherwise
 *
 * This private function allows to lock a socket for private use; this
 * prevents others thread from using it.
 *
 * Locking the socket is a blocking operation, so it should be avoided
 * whenever possible.
 *
 * @Warning
 * Do not assume that you own the socket when this function returns; you
 * MUST check the return value to ensure the lock is effective.
 *
 * When a socket is on the verge of destruction, its lock is broken, which
 * cause all pending socket_lock() calls to return -1. If you foolishly
 * ignore this return value, you may very well end up using a dead socket.
 *
 * Once you are done with the socket, you MUST unlock it to allow other
 * threads to access it.
 *
 * @see socket_unlock()
 *
 */

/* -------------------------------------------------------------------------- */

private void socket_unlock(m_socket *s);

/**
 * @ingroup socket
 * @fn void socket_unlock(m_socket *s)
 * @param s the socket to unlock
 * @return void
 *
 * This function unconditionnaly unlock the socket, silently handling broken
 * locks. Once the socket has been unlocked, it will be accessible to other
 * threads back again.
 *
 * If you call this function more than once, as much socket_lock() calls
 * will succeed. This may be useful to implement complex locking schemes.
 *
 * @see socket_lock()
 *
 */

/* -------------------------------------------------------------------------- */

public int socket_connect(m_socket *s);

/**
 * @ingroup socket
 * @fn int socket_connect(m_socket *s)
 * @param s the socket to connect
 * @return 0 if the connection was successful, -1 otherwise
 *
 * This function connects the socket to the host and port with which it has
 * been initialized.
 *
 * If the socket is non blocking (default), it is not sufficient to call
 * this function for the connection to be established; the Mammouth server
 * monitors connections in progress and notify their master plugins.
 * So, if you call this function from a plugin, using non blocking i/o,
 * you should wait for it.
 *
 * If you use blocking i/o, the connection will be useable immediately
 * after the function returns 0.
 *
 */

/* -------------------------------------------------------------------------- */

private ssize_t socket_write_lowlevel(m_socket *s, const char *data,
                                      size_t len                    );

/**
 * @ingroup socket
 * @fn ssize_t socket_write_lowlevel(m_socket *s, const char *data, size_t len)
 * @param s the socket
 * @param data the data to send
 * @param len the lenght of the data
 * @return specific error codes, see below
 *
 * @note This is a private function, it should not be called from a plugin.
 *
 * This function do all the dirty work to send the data over the wire. It
 * may return several error codes:
 * > 0: the returned number of bytes were successfully written
 * SOCKET_ECLOSE: the connection is down
 * SOCKET_EAGAIN: recoverable error, retry the call
 * SOCKET_EFATAL: fatal error
 *
 */

/* -------------------------------------------------------------------------- */

private ssize_t socket_read_lowlevel(m_socket *s, char *out, size_t len);

/**
 * @ingroup socket
 * @fn ssize_t socket_read_lowlevel(m_socket *s, char *out, size_t len)
 * @param s the socket
 * @param out the buffer to copy the data read in
 * @param len the size of the buffer
 * @return specific error code, see below
 *
 * @note This is a private function, it should not be called from a plugin.
 *
 * This function reads the incoming data from the socket and copy them
 * to the given buffer, up to its length.
 *
 * Like @ref socket_write_lowlevel(), this function returns various
 * error codes:
 * > 0: the returned number of bytes were successfully read and copied to the
 *      buffer
 * SOCKET_ECLOSE: the connection is down
 * SOCKET_EAGAIN: recoverable error, retry the call
 * SOCKET_EFATAL: fatal error
 *
 */

/* -------------------------------------------------------------------------- */

public m_socket *socket_close(m_socket *s);

/**
 * @ingroup socket
 * @fn m_socket *socket_close(m_socket *s)
 * @param s the socket to destroy
 * @return always NULL
 *
 * This function gracefully shutdown the connection and destroy the associated
 * socket structure.
 *
 * You may clean up the pointer with this function, as it always return NULL.
 *
 */

/* -------------------------------------------------------------------------- */

#endif
