#include "m_string.h"

#define _UINT(c) ((unsigned int)((unsigned char)(c)))

#define _STRING_FLAG_FIXLEN 0x01 /* disable string resizing */
#define _STRING_FLAG_RDONLY 0x02 /* disable string writing */
#define _STRING_FLAG_STATIC 0x03 /* disable writing and resizing */

/* -- base64 encoding/decoding specific variables -- */

/* encoding lookup table */
static char _b64[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                     "abcdefghijklmnopqrstuvwxyz"
                     "0123456789+/";

/* decoding lookup table */
static char _d64[] = {
                      -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /*  12 */
                      -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /*  24 */
                      -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, /*  36 */
                      -1, -1, -1, -1, -1, -1, -1, 62, -1, -1, -1, 63, /*  48 */
                      52, 53, 54, 55, 56, 57, 58, 59, 60, 61, -1, -1, /*  60 */
                      -1, -1, -1, -1, -1,  0,  1,  2,  3,  4,  5,  6, /*  72 */
                       7,  8, 9,  10, 11, 12, 13, 14, 15, 16, 17, 18, /*  84 */
                      19, 20, 21, 22, 23, 24, 25, -1, -1, -1, -1, -1, /*  96 */
                      -1, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, /* 108 */
                      37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, /* 120 */
                      49, 50, 51, -1, -1, -1, -1, -1                  /* 128 */
                     };

/* macro to ease the use of the decode table and prevent out of bound access */
#define _D64(c) (_d64[(c) & 0x7F])

/* -------------------------------------------------------------------------- */
/* 1. Basic string allocation, conversion and initialization functions        */
/* -------------------------------------------------------------------------- */

public m_string *string_alloc(const char *string, size_t len)
{
    /** @brief allocate a m_string and initialize it with the given data */

    m_string *new = malloc(sizeof(*new));

    if (! new) { perror(ERR(string_alloc, malloc)); return NULL; }

    new->_data = NULL;

    /* allocate the internal buffer */
    if (len) {
        new->_data = malloc( (len + 1) * sizeof(*new->_data));
        if (! new->_data) {
            perror(ERR(string_alloc, malloc)); free(new); return NULL;
        }
    
        /* copy the given data in the internal buffer */
        if (string) memmove(new->_data, string, len);

        new->_data[len] = '\0';
    }

    new->_len = len; new->_alloc = len; new->_flag = 0;
    new->parts = 0; new->token = NULL;

    return new;
}

/* -------------------------------------------------------------------------- */

public format(2, 3) m_string *string_fmt(m_string *str, const char *fmt, ...)
{
    /** @brief overwrite a string with the given format or allocate it */

    va_list args;
    int ret = 0;
    m_string *buffer = NULL;

    /* sanity check */
    if (! fmt || (str && str->_flag & _STRING_FLAG_STATIC) ) return NULL;

    /* allocate a buffer of arbitrary sizeof(int) size */
    buffer = string_alloc(NULL, sizeof(int));

    /* this algorithm come from the GNU vsnprintf man page ;) */
    while (1) {
        /* try to print in the allocated space */
        va_start(args, fmt);
        ret = vsnprintf(buffer->_data, buffer->_alloc, fmt, args);
        va_end(args);

        if ( (ret > -1) && ((size_t) ret < buffer->_alloc) ) {
            /* if no string was given, allocate a clean one */
            if (! str) str = string_alloc(NULL, 0);
            free(str->_data); string_free_token(str);
            
            /* fill the string with buffer infos and return it */
            str->_len = ret; str->_alloc = buffer->_alloc;
            str->_data = buffer->_data; buffer->_data = NULL;
            string_free(buffer);
            
            return str;
        }
        
        /* else try again with more space */
        if (ret > -1) {
            /* glibc 2.1 - precisely what is needed */
            if (string_extend(buffer, ret + 1) == -1)
                return string_free(buffer);
        } else {
            /* glibc 2.0 - try twice the old size */
            if (string_extend(buffer, buffer->_alloc * 2) == -1)
                return string_free(buffer);
        }
    }

    return NULL;
}

/* -------------------------------------------------------------------------- */

public inline m_string *string_from_uint32(uint32_t u32)
{
    /** @brief create a string from an integer */

    return string_alloc((char *) & u32, sizeof(u32));
}

/* -------------------------------------------------------------------------- */

public inline int string_to_uint32(const m_string *string, uint32_t *out)
{
    /** @brief convert a binary string to an integer */

    if (! string || ! out || ! CSTR(string) || SIZE(string) < sizeof(*out))
        return -1;

    memcpy(out, string->_data, sizeof(*out));

    return 0;
}

/* -------------------------------------------------------------------------- */

public inline m_string *string_from_uint64(uint64_t u64)
{
    /** @brief create a string from an integer */

    return string_alloc((char *) & u64, sizeof(u64));
}

/* -------------------------------------------------------------------------- */

public inline int string_to_uint64(const m_string *string, uint64_t *out)
{
    /** @brief convert a binary string to an integer */

    if (! string || ! out || ! CSTR(string) || SIZE(string) < sizeof(*out))
        return -1;

    memcpy(out, string->_data, sizeof(*out));

    return 0;
}

/* -------------------------------------------------------------------------- */

public int string_wchar(m_string *string)
{
    /** @brief convert a multibyte string to a wide character string */

    char *buffer = NULL;
    size_t bufsize = 0;

    if (! string || ! CSTR(string)) return -1;

    /* check if writing and resizing to the string is allowed */
    if (string->_flag & _STRING_FLAG_STATIC) return -1;

    /* find the size of the wchar string */
    if ( (bufsize = mbstowcs(NULL, CSTR(string), 0)) != (size_t) -1) {
        /* allocate the wchar buffer */
        buffer = malloc( (bufsize + 1) * sizeof(wchar_t));
        if (buffer == NULL) {
            perror(ERR(string_wchar, malloc)); return -1;
        }

        /* convert the multibyte string to wchar */
        if (mbstowcs((wchar_t *) buffer, CSTR(string), bufsize + 1) == (size_t) -1) {
            free(buffer); perror(ERR(string_wchar, mbstowcs));
            return -1;
        }

        /* update the string */
        free(string->_data); string->_data = buffer;
        string->_len = bufsize * sizeof(wchar_t);
        string->_alloc = (bufsize + 1) * sizeof(wchar_t);
    } else {
        perror(ERR(string_wchar, mbstowcs)); return -1;
    }

    return 0;
}

/* -------------------------------------------------------------------------- */

public int string_mbyte(m_string *string)
{
    /** @brief convert a wide character string to a multibyte string */

    char *buffer = NULL;
    size_t bufsize = 0;

    if (! string || ! CSTR(string)) return -1;

    /* check if writing and resizing to the string is allowed */
    if (string->_flag & _STRING_FLAG_STATIC) return -1;

    /* find the size of the multibyte buffer */
    if ( (bufsize = wcstombs(NULL, WSTR(string), 0)) != (size_t) -1) {
        /* allocate it */
        buffer = malloc( (bufsize + 1) * sizeof(*buffer));
        if (buffer == NULL) {
            perror(ERR(string_mbyte, malloc)); return -1;
        }

        /* perform the conversion */
        if (wcstombs(buffer, WSTR(string), bufsize + 1) == (size_t) -1) {
            perror(ERR(string_mbyte, wcstombs));
            free(buffer); return -1;
        }

        /* all went fine, update the string */
        free(string->_data); string->_data = buffer;
        string->_len = bufsize; string->_alloc = bufsize + 1;
    } else {
        perror(ERR(string_mbyte, wcstombs)); return -1;
    }

    return 0;
}
            
/* -------------------------------------------------------------------------- */

public int string_swap(m_string *string)
{
    /** @brief swap a binary string */

    unsigned int i = 0;
    unsigned char tmp = 0;

    if (! string || ! CSTR(string) || SIZE(string) % 2) return -1;

    for (i = 0; i < SIZE(string); i += 2) {
        tmp = string->_data[i];
        string->_data[i] = string->_data[SIZE(string) - 1 - i];
        string->_data[SIZE(string) - 1 - i] = tmp;
    }

    string_free_token(string);

    return 0;
}

/* -------------------------------------------------------------------------- */

public inline m_string *string_dup(const m_string *string)
{
    /** @brief duplicate a string */

    return string ? string_alloc(CSTR(string), SIZE(string)) : NULL;
}

/* -------------------------------------------------------------------------- */

public inline char *string_dups(const char *string, size_t len)
{
    /** @brief duplicate a C string */

    char *ret = NULL;

    if (! string || ! len) return NULL;

    ret = malloc( (len + 1) * sizeof(*ret));
    if (! ret) { perror(ERR(string_dups, malloc)); return NULL; }

    memcpy(ret, string, len); ret[len] = '\0';

    return ret;
}

/* -------------------------------------------------------------------------- */

public m_string *string_free(m_string *string)
{
    /** @brief clean up a m_string and its internal buffer */

    if (string) { free(string->_data); free(string->token); } free(string);

    return NULL;
}

/* -------------------------------------------------------------------------- */
/* 2. String size functions                                                   */
/* -------------------------------------------------------------------------- */

public size_t string_len(const m_string *string)
{
    /** @brief return the length of the string data */

    return string ? string->_len : (size_t) -1;
}

/* -------------------------------------------------------------------------- */

public size_t string_spc(const m_string *string)
{
    /** @brief return the string allocated buffer size */

    return string ? string->_alloc : (size_t) -1;
}

/* -------------------------------------------------------------------------- */

public int string_dim(m_string *string, size_t size)
{
    /** @brief force resize the internal buffer of a string */

    char *new = NULL;

    /* sanity checks */
    if (! string || ! size) return -1;
    if (string->_alloc == size) return 0;

    /* do not resize a fixed length string */
    if (string->_flag & _STRING_FLAG_FIXLEN) return -1;

    /* resize the internal buffer */
    new = realloc(string->_data, (size + 1) * sizeof(*string->_data));
    if (! new) { perror(ERR(string_dim, realloc)); return -1; }

    /* all went fine, update the string */
    string->_data = new; string->_alloc = size;
    string->_data[size] = '\0';
    string->_len = MIN(SIZE(string), size);

    /* tokens are likely corrupted, so trash them */
    string_free_token(string);

    return 0;
}

/* -------------------------------------------------------------------------- */

public inline int string_extend(m_string *string, size_t size)
{
    /** @brief resize a string only if the new size is greater */

    if (! string || ! size) return -1;

    return (size > string->_alloc) ? string_dim(string, size) : 0;
}

/* -------------------------------------------------------------------------- */

public inline int string_shrink(m_string *string, size_t size)
{
    /** @brief resize a string only if the new size is lower */

    if (! string || ! size) return -1;

    return (size < string->_alloc) ? string_dim(string, size) : 0;
}

/* -------------------------------------------------------------------------- */
/* 3. Basic string operations (char * flavor)                                 */
/* -------------------------------------------------------------------------- */

public m_string *string_movs(m_string *to, off_t o, const char *from, size_t l)
{
    /** @brief move a C string or a part of it into a m_string */

    int within = 0; ptrdiff_t within_offset = 0;

    if (o >= 0 && from && l) {
        /* if to does not exist, allocate it */
        if ( (to == NULL) && ( (to = string_alloc(NULL, 0)) == NULL) )
            return NULL;

        /* check if writing to the string is allowed */
        if (to->_flag & _STRING_FLAG_RDONLY) return NULL;

        /* check if the source is inside the destination */
        if ( (from >= CSTR(to)) && (from < CSTR(to) + to->_alloc) )
            within = 1, within_offset = from - CSTR(to);

        /* reject out of bound offsets and resize the destination string */
        if (string_extend(to, o + l) == -1) {
            return NULL;
        } else if (within) {
            /* if the source was within destination, correct the pointer */
            from = CSTR(to) + within_offset;

            /* perform a sanity check to see if the offset is still in bound */
            if ( (from < CSTR(to)) || (from >= CSTR(to) + SIZE(to)) )
                return NULL;
        }

        memmove(to->_data + o, from, l);

        /* all went fine, update the string */
        to->_len = (SIZE(to) < o + l) ? o + l : SIZE(to);
        to->_data[SIZE(to)] = '\0';

        /* if there were some tokens, they are likely corrupted... drop them */
        string_free_token(to);
    }

    return to;
}

/* -------------------------------------------------------------------------- */

public inline m_string *string_cats(m_string *to, const char *from, size_t len)
{
    /** @brief append a C string to a m_string */

    if (to == NULL) return NULL;

    return string_movs(to, SIZE(to), from, len);
}

/* -------------------------------------------------------------------------- */

public inline m_string *string_pres(m_string *to, const char *from, size_t len)
{
    /** @brief prepend a C string to a m_string */

    /* move the original string to the right */
    if ( (to != NULL) && (string_movs(to, len, CSTR(to), to->_len) == NULL) ) {
        return NULL;
    }

    return string_movs(to, 0, from, len);
}

/* -------------------------------------------------------------------------- */

public inline m_string *string_cpys(m_string *to, const char *from, size_t len)
{
    /** @brief copy data from a C string to a m_string */

    return string_movs(to, 0, from, len);
}

/* -------------------------------------------------------------------------- */

public int string_cmps(const m_string *a, const char *b, size_t len)
{
    /** @brief compare a m_string and a C string */

    if (! a || ! CSTR(a) || ! b || ! len) return 2;

    return memcmp(CSTR(a), b, MIN(SIZE(a), len));
}

/* -------------------------------------------------------------------------- */
/* 4. Basic string operations (m_string flavor)                               */
/* -------------------------------------------------------------------------- */

public inline m_string *string_mov(m_string *to, off_t off,
                                   const m_string *from, size_t len)
{
    /** @brief move a string or a part of it into another string */

    if (! from) return NULL;

    return string_movs(to, off, CSTR(from), len);
}

/* -------------------------------------------------------------------------- */

public inline m_string *string_cat(m_string *to, const m_string *from)
{
    /** @brief append a string to another */

    if (from == NULL) return NULL;

    return string_cats(to, CSTR(from), SIZE(from));
}

/* -------------------------------------------------------------------------- */

public inline m_string *string_pre(m_string *to, const m_string *from)
{
    /** @brief prepend a string to another */

    if (from == NULL) return NULL;

    return string_pres(to, CSTR(from), SIZE(from));
}

/* -------------------------------------------------------------------------- */

public inline m_string *string_cpy(m_string *to, const m_string *from)
{
    /** @brief copy data from a string to another */

    if (from == NULL) return NULL;

    return string_cpys(to, CSTR(from), SIZE(from));
}

/* -------------------------------------------------------------------------- */

public inline int string_cmp(const m_string *a, const m_string *b)
{
    /** @brief compare two strings */

    if (b == NULL) return 2;

    return string_cmps(a, CSTR(b), SIZE(b));
}

/* -------------------------------------------------------------------------- */

public int string_upper(m_string *string)
{
    /** @brief convert an ASCII string to upper case */

    char *base = NULL;

    if (string == NULL || CSTR(string) == NULL) return -1;

    /* check if writing to the string is allowed */
    if (string->_flag & _STRING_FLAG_RDONLY) return -1;

    base = string->_data;

    while ( (*string->_data = toupper(*string->_data ++)) );

    string->_data = base;

    /* since it is an in place transformation, keep existings tokens */

    return 0;
}

/* -------------------------------------------------------------------------- */

public int string_lower(m_string *string)
{
    /** @brief convert an ASCII string to lower case */

    char *base = NULL;

    if (! string || ! CSTR(string)) return -1;

    /* check if writing to the string is allowed */
    if (string->_flag & _STRING_FLAG_RDONLY) return -1;

    base = string->_data;

    while ( (*string->_data = tolower(*string->_data ++)) );

    string->_data = base;

    /* since it is an in place transformation, keep existings tokens */

    return 0;
}

/* -------------------------------------------------------------------------- */
/* 5. Advanced string manipulation functions                                  */
/* -------------------------------------------------------------------------- */

public off_t string_finds(const m_string *str, size_t o, const char *sub,
                          size_t len                                     )
{
    /** @brief find a C susbstring within a string */

    int lut[UCHAR_MAX + 1];
    size_t shift = 0, i = 0, k = 0;

    /* sanity checks */
    if (unlikely(! str || ! CSTR(str) || ! sub || ! len)) return -1;
    if (unlikely( (len > SIZE(str)) || (o > SIZE(str)) )) return -1;

    /* register the distance between each possible byte and the last byte
       of the substring in a lookup table */
    for (i = 0; i < UCHAR_MAX + 1; i ++) lut[i] = len;

    /*  fix the length */
    len --;
    
    for (i = 0; i < len; i ++) lut[_UINT(sub[i])] -= i + 1;

    /* get the shift length and set up the guard char */
    shift = lut[_UINT(sub[len])]; lut[_UINT(sub[len])] = 0;

    /* search loop (Tuned Boyer-Moore algorithm) */
    for (i = o; i < SIZE(str); i += shift) {
        /* use the lookup table to skip impossible matches */
        do {
            register size_t l = i + len;
            if (likely(l < SIZE(str)))
                k = lut[_UINT(str->_data[l])];
            else
                k = 0;
        } while ( (k != 0) && (i += k) );
        /* (no manual unrolling, GCC will take care of that for us ;) */

        /* try a possible match */
        if (likely(i < SIZE(str)) && memcmp(sub, CSTR(str) + i, len) == 0) break;
    }
    
    return (likely(i >= SIZE(str))) ? (size_t) -1 : i;
}

/* -------------------------------------------------------------------------- */

public inline off_t string_find(const m_string *string, const m_string *sub)
{
    /** @brief find a susbstring within a string */

    if (! string || ! sub || ! CSTR(string) || ! CSTR(sub) ) return -1;

    return string_finds(string, 0, CSTR(sub), SIZE(sub));
}

/* -------------------------------------------------------------------------- */

public int string_splits(m_string *string, const char *pattern, size_t len)
{
    size_t *offset = NULL, off = 0;
    unsigned int i = 0, count = 0;
    m_string *token = NULL;

    if (! string || ! CSTR(string) || ! pattern || ! len) return -1;

    /* maximum number of tokens */
    count = SIZE(string) / len + 1;

    if ( (offset = malloc(count * sizeof(*offset))) == NULL) goto _error_malloc;

    /* fill the offsets array with the locations of the pattern */
    while ( (offset[i] = string_finds(string, off, pattern, len)) != (size_t) -1) {
        off = offset[i] + len; if (i < count) i ++; else break;
    }

    if ( (count = i) == 0) goto _error_notfnd; else count ++;

    if ( (token = malloc(count * sizeof(*token))) == NULL) goto _error_malloc;

    for (i = 0, off = 0; i < count; off = offset[i] + len, i ++) {
        token[i]._flag = _STRING_FLAG_STATIC;
        token[i]._data = string->_data + off;
        token[i]._len = (offset[i] != (size_t) -1) ? offset[i] : SIZE(string);
        token[i]._len -= off;
        token[i]._alloc = -1; token[i].parts = 0; token[i].token = NULL;
    }

    free(offset);

    string_free_token(string); string->parts = count; string->token = token;

    return 0;

_error_malloc: perror(ERR(string_splits, malloc));
_error_notfnd: free(offset);
               return -1;
}

/* -------------------------------------------------------------------------- */

public inline int string_split(m_string *string, const m_string *pattern)
{
    if (! string || ! pattern || ! CSTR(pattern)) return -1;

    return string_splits(string, CSTR(pattern), SIZE(pattern));
}

/* -------------------------------------------------------------------------- */

public m_string *string_reps(m_string *string, const char *search, size_t slen,
                             const char *rep, size_t rlen                      )
{
    /** @brief replace all occurences of a C substring in a string */

    size_t count = 0;
    off_t *offset = NULL, off = 0, src = 0, dst = 0;
    m_string *buf = NULL;
    unsigned int i = 0;

    /* a NULL replacement string is allowed (deletion) */
    if (! string || ! CSTR(string) || ! search || ! slen) return NULL;

    /* check if writing to the string and resizing is allowed */
    if (string->_flag & _STRING_FLAG_STATIC) return NULL;

    /* find the maximal number of replacement and allocate the offset array */
    count = SIZE(string) / slen + 1; offset = malloc(count * sizeof(*offset));
    if (offset == NULL) { perror(ERR(string_reps, malloc)); return NULL; }

    /* fill the offset array with the location of the search strings */
    while ( (offset[i] = string_finds(string, off, search, slen)) != -1) {
        off = offset[i] + slen; if (i < count) i ++; else break;
    }

    if ( (count = i) == 0) goto _error_nfnd;

    /* find the size of the buffer after the replacement and allocate it */
    if ( ! (buf = string_alloc(NULL, SIZE(string) + count * (rlen - slen))) )
        goto _error_buff;

    /* perform the replacement */
    for (i = 0; i < count; i ++) {
        /* copy the data in-between the strings to replace */
        if (! string_movs(buf, dst, CSTR(string) + src, offset[i] - src))
            goto _error_move;
        dst += offset[i] - src; src = offset[i] + slen;

        /* loop no further without proper replacement string */
        if (! rep || ! rlen) continue;

        /* replace the string */
        if (! string_movs(buf, dst, rep, rlen))
            goto _error_move;
        dst += rlen;
    }

    free(string->_data); string->_data = buf->_data; buf->_data = NULL;
    string->_len = buf->_len; string->_alloc = buf->_alloc;
    string_free_token(string); string_free(buf); free(offset);

    return string;

_error_move: string_free(buf);
_error_nfnd: /* not found */
_error_buff: free(offset);
             return NULL;
}

/* -------------------------------------------------------------------------- */

public m_string *string_rep(m_string *string, const m_string *search,
                            const char *rep, size_t rlen             )
{
    /** @brief replace all occurences of a substring in a string */

    if (! string || ! search || ! CSTR(search)) return NULL;

    /* check if writing to the string and resizing is allowed */
    if (string->_flag & _STRING_FLAG_STATIC) return NULL;

    return string_reps(string, CSTR(search), SIZE(search), rep, rlen);
}

/* -------------------------------------------------------------------------- */

public inline m_string *string_rems(m_string *str, const char *rem, size_t len)
{
    /** @brief remove all occurences of a C substring in a string */

    return string_reps(str, rem, len, NULL, 0);
}

/* -------------------------------------------------------------------------- */

public inline m_string *string_rem(m_string *string, const m_string *rem)
{
    /** @brief remove all occurences of a substring in a string */

    if (! string || ! CSTR(string) || ! rem || ! CSTR(rem)) return NULL;

    return string_reps(string, CSTR(rem), SIZE(rem), NULL, 0);
}

/* -------------------------------------------------------------------------- */

public m_string *string_b64s(const char *s, size_t size, size_t linesize)
{
    /** @brief convert a C string to base64 encoding */

    size_t i = 0, j = 0, crlf = 0, b64size = 0;
    char *ret = NULL;
    m_string *sret = NULL;

    if (unlikely(! s || ! size)) return NULL;

    /* reject bogus linesize */
    if ( unlikely( (linesize) && ((linesize % 4) || (linesize > 72)) ) )
        return NULL;

    /* compute the size of the base64 encoded string, with a trailing NUL */
    b64size = (((size + 3 - (size % 3)) / 3) * 4) + 1;

    /* add the size of CRLF depending on the linesize */
    if (linesize) b64size += ((b64size / linesize) * 2) + 2;

    if ( unlikely( ( (ret = calloc(b64size, sizeof(*ret))) == NULL ) ) )
        return NULL;

    while (i + 3 < size) {

        if (likely(j + 4 < b64size)) {
            /* convert three bytes in ASCII characters */
            ret[j ++] = _b64[(s[i] & 0xFC) >> 2];
            ret[j ++] = _b64[((s[i] & 0x03) << 4) | ((s[i + 1] & 0xF0) >> 4)];
            ret[j ++] = _b64[((s[i + 1] & 0x0F) << 2) | ((s[i + 2] & 0xC0) >> 6)];
            ret[j ++] = _b64[s[i + 2] & 0x3F]; i += 3;
        } else goto _panic;

        /* if a linesize was given, check if we should insert a CRLF */
        if (linesize && ((j - crlf) % linesize == 0) ) {
            if (likely(j + 2 < b64size)) {
                ret[j ++] = '\r'; ret[j ++] = '\n'; crlf += 2;
            } else goto _panic;
        }
    }

    if ((size + 1 - i) && likely(j + 4 < b64size)) {
        /* add some padding */
        memset(ret + j, '=', 4);
            
        /* process the remaining */
        switch (size + 1 - i) {
        case 4:
        ret[j + 3] = _b64[s[i + 2] & 0x3F];
        case 3:
        ret[j + 2] = _b64[(s[i + 1] & 0x0F) << 2 | (s[i + 2] & 0xC0) >> 6];
        case 2:
        ret[j + 1] = _b64[(s[i] & 0x03) << 4 | (s[i + 1] & 0xF0) >> 4];
        case 1:
        ret[j] = _b64[(s[i] & 0xFC) >> 2]; j += 4;
        }
    }

    /* add the last CRLF if formatting was asked */
    if ( (linesize) && likely(j + 2 < b64size) ) {
        ret[j] = '\r'; ret[j + 1] = '\n';
    }

    /* package the buffer in a string_m */
    if ( (sret = string_alloc(NULL, 0)) == NULL) {
        free(ret);
    } else {
        sret->_flag = 0; sret->_data = ret;
        sret->_len = b64size - 1; sret->_alloc = b64size -1;
        sret->parts = 0; sret->token = NULL;
    }

    return sret;

_panic:
    free(ret);
    return NULL;
}

/* -------------------------------------------------------------------------- */

public inline m_string *string_b64(const m_string *s, size_t linesize)
{
    /** @brief convert a string to base64 encoding */

    return string_b64s(CSTR(s), SIZE(s), linesize);
}

/* -------------------------------------------------------------------------- */

public m_string *string_deb64s(const char *s, size_t size)
{
    /** @brief convert a base64 encoded C string to plain text */

    char *ret = NULL;
    size_t i = 0, j = 0;
    m_string *sret = NULL;

    if (unlikely(! s || ! size)) return NULL;

    /* don't know the number of CRLF or other junk, so allocate the same size */
    if ( unlikely( ( (ret = calloc(size + 1, sizeof(*ret))) == NULL ) ) )
        return NULL;
    
    /* discard trailing garbage */
    while (-- size && _D64(s[size]) == -1);

    while (i + 4 < size) {
        if (likely(_D64(s[i]) != -1)) {
            /* take four ASCII chars and convert them in three bytes */
            ret[j ++] = _D64(s[i]) << 2 | _D64(s[i + 1]) >> 4;
            ret[j ++] = _D64(s[i + 1]) << 4 | _D64(s[i + 2]) >> 2;
            ret[j ++] = ((_D64(s[i + 2]) << 6) & 0xC0) | _D64(s[i + 3]);
            i += 4;
        } else i ++; /* drop CRLF and other noise */
    }

    /* skip garbage between the last processed piece and the remaining */
    while (_D64(s[i]) == -1 && i ++ < size);

    /* process the remaining */
    switch (size - i) {
    case 3:
    ret[j + 2] = ((_D64(s[i + 2]) << 6) & 0xC0) | _D64(s[i + 3]);
    case 2:
    ret[j + 1] = _D64(s[i + 1]) << 4 | _D64(s[i + 2]) >> 2;
    case 1:
    ret[j] = _D64(s[i]) << 2 | _D64(s[i + 1]) >> 4; j += 3;
    }

    /* package the buffer in a m_string */
    if ( (sret = string_alloc(NULL, 0)) == NULL) {
        free(ret);
    } else {
        sret->_flag = 0; sret->_data = ret;
        sret->_len = j; sret->_alloc = i;
        sret->parts = 0; sret->token = NULL;
    }

    return sret;
}

/* -------------------------------------------------------------------------- */

public inline m_string *string_deb64(const m_string *s)
{
    /** @brief convert a base64 encoded string to plain text */

    return string_deb64s(CSTR(s), SIZE(s));
}

/* -------------------------------------------------------------------------- */

public void string_free_token(m_string *string)
{
    unsigned int i = 0;

    if (string) {
        /* recursively clean the tokens' tokens, if any */
        for (i = 0; i < PARTS(string); i ++)
            string_free_token(& string->token[i]);

        free(string->token); string->token = NULL; string->parts = 0;
    }
}

/* -------------------------------------------------------------------------- */
