#ifndef M_STRING_H

#define M_STRING_H

#include "m_core_def.h"

/** @defgroup string core::string */

typedef struct m_string {
    /* private */
    unsigned int _flag;
    size_t _len;
    size_t _alloc;
    char *_data;

    /* public */
    unsigned int parts;
    struct m_string *token;
} m_string;

/**
 * @ingroup string
 * @struct m_string
 *
 * This structure is designed to hold arbitrary sequences of bytes. Its fields
 * should never be directly accessed, at the exception of the @ref parts
 * and @ref token fields, which are considered public members.
 *
 * @ref parts holds the number of token in which the string has been sliced,
 * with @ref string_split or whatever function.
 *
 * @ref token is an array of m_string structures, which are set to points
 * to particular sub sections of the current string. Those tokens are "static",
 * this means they can not be written into, nor resized. These restrictions
 * are here to avoid unvolontary corruptions of the main string while messing
 * with its tokens.
 * However, tokens can be used perfectly fine with all functions which accept
 * read-only strings, like standard m_string structures.
 * In all case, functions modifying their string parameters will voluntarily
 * fail when encountering any "static" strings, including tokens.
 *
 * Even if these members are public, the macros @ref TOKEN and @ref PARTS have
 * been defined to use them more conveniently.
 *
 * You can clean up the tokens of a string at any time by calling
 * @ref string_free_token().
 *
 * All other fields are private, and accessing them directly may cause your
 * program to crash or your hairs to fall...
 *
 * This being said, here is their descriptions.
 *
 * @b private @ref _flag is a bitfield holding several internal flags, like
 * write protection, resizing protection...
 *
 * @b private @ref _len is the actual size, in bytes, of the data stored inside
 * the m_string structure. For obvious reasons, messing with it may not be a
 * great idea...
 *
 * @b private @ref _alloc is the actual size of the inner data buffer of the
 * m_string structure. It is mainly used in the resizing functions.
 *
 * @b private @ref _data is the internal buffer where the data are stored.
 * If you want to mess with them directly, why wrap them in a structure anyway.
 *
 */

/** Macro for accessing a token inside a m_string. No bound check, be careful */
#define TOKEN(s, i) ((const m_string *) & (s)->token[(i)])
/** Macro for accessing the number of token held inside a m_string */
#define PARTS(s)    ((s)->parts)

/** Macro to READ-ONLY access to the inner data buffer of a m_string */
#define CSTR(s) ((const char *) (s)->_data)
/** Macro to READ-ONLY access to the inner data buffer, wchar_t flavor */
#define WSTR(s) ((const wchar_t *) (s)->_data)
/** Macro to get the size of a string. Consider using @ref string_len instead */
#define CLEN(s) ((s)->_len)
/** Macro to get the size of a string, wchar_t flavor */
#define WLEN(s) ((s)->_len / sizeof(wchar_t))
/** Prettier alias to CLEN() */
#define SIZE    CLEN

/* -------------------------------------------------------------------------- */

public m_string *string_alloc(const char *string, size_t len);

/** string_allocw is an alias to conveniently allocate wide char string */
#define string_allocw(w, l) \
        (string_alloc((char *) (w), (l) * sizeof(wchar_t)))

/**
 * @ingroup string
 * @fn m_string *string_alloc(const char *string, size_t len)
 * @param string the data with which the new string will be initialized.
 * @param len the size of the buffer to be allocated.
 * @return a pointer to a new m_string, or NULL.
 *
 * This function allocates a new m_string structure, and its internal buffer if
 * the @b len parameter is valid (non zero).
 *
 * If a pointer to a @b string was given, up to @b len bytes of it are copied
 * into the freshly allocated buffer.
 *
 * So, it is perfectly legal to do:
 * my_string = string_alloc(NULL, 256); to allocate a 256 bytes wide buffer
 * or:
 * my_string = string_alloc(NULL, 0) to allocate an empty m_string structure.
 *
 * However, if a pointer to some data are provided, the @b len parameter must
 * be coherent, or the function will fail.
 *
 * A m_string structure should be destroyed with @ref string_free() after use.
 *
 */

/* -------------------------------------------------------------------------- */

public format(2, 3) m_string *string_fmt(m_string *str, const char *fmt, ...);

/**
 * @ingroup string
 * @fn m_string *string_fmt(m_string *str, const char *fmt, ...)
 * @param str the string to be overwritten, may be NULL.
 * @param fmt the printf(3) compatible format.
 * @param ... the arguments matching the format.
 * @return NULL if an error occured, a pointer to the @b str string otherwise.
 *
 * This function overwrites the given string (or a newly allocated one if the
 * @b str parameter was NULL) with the given formatted output.
 *
 * Since this function allocates a string and returns it if the @b str
 * parameter is NULL, it is recommended to always check its return value to
 * avoid memory leaks.
 *
 * If an error occurs, the original string will be left unchanged and the
 * function will return NULL.
 *
 */

/* -------------------------------------------------------------------------- */

public inline m_string *string_from_uint32(uint32_t u32);

/**
 * @ingroup string
 * @fn m_string *string_from_uint32(uint32_t u32)
 * @param u32 the integer to be converted into a binary string
 * @return NULL if the conversion failed, a pointer to a new string otherwise
 *
 * This function creates a new m_string from a 32 bits integer. The resulting
 * binary string is not swapped.
 *
 */

/* -------------------------------------------------------------------------- */

public inline int string_to_uint32(const m_string *string, uint32_t *out);

/**
 * @ingroup string
 * @fn int string_to_uint32(const m_string *string, uint32_t *out)
 * @param string the binary string to convert to a 32 bits integer
 * @param out a pointer to a 32 bits integer for storage
 * @return -1 if an error occured, 0 otherwise
 *
 * This function simply converts a binary string into a 32 bits integer,
 * without swapping.
 *
 * If an error occurs, the function will return -1.
 *
 */

/* -------------------------------------------------------------------------- */

public inline m_string *string_from_uint64(uint64_t u64);

/**
 * @ingroup string
 * @fn m_string *string_from_uint64(uint64_t u64)
 * @param u64 the integer to be converted into a binary string
 * @return NULL if the conversion failed, a pointer to a new string otherwise
 *
 * This function creates a new m_string from a 64 bits integer. The resulting
 * binary string is not swapped.
 *
 */

/* -------------------------------------------------------------------------- */

public inline int string_to_uint64(const m_string *string, uint64_t *out);

/**
 * @ingroup string
 * @fn int string_to_uint64(const m_string *string, uint64_t *out)
 * @param string the binary string to convert to a 64 bits integer
 * @param out a pointer to a 64 bits integer for storage
 * @return -1 if an error occured, 0 otherwise
 *
 * This function simply converts a binary string into a 64 bits integer,
 * without swapping.
 *
 * If an error occurs, the function will return -1.
 *
 */

/* -------------------------------------------------------------------------- */

public int string_wchar(m_string *string);

/**
 * @ingroup string
 * @fn int string_wchar(m_string *string)
 * @param string the string to be converted.
 * @return -1 if an error occured, 0 otherwise.
 *
 * This function converts the data stored in the internal buffer of the given
 * @b string to wide characters from multibyte characters.
 *
 * Of course, for this conversion to make sense, the data stored should better
 * be of textual nature.
 *
 * If the conversion is successfull, the inner buffer is replaced by its
 * multibyte equivalent.
 *
 * If the fonction fails to convert the data, it will returns -1 and leave the
 * buffer unchanged.
 *
 */

/* -------------------------------------------------------------------------- */

public int string_mbyte(m_string *string);

/**
 * @ingroup string
 * @fn int string_mbyte(m_string *string)
 * @param string the string to be converted.
 * @return -1 if an error occured, 0 otherwise.
 *
 * This function converts the data stored in the internal buffer of the given
 * @b string to multibyte characters from wide characters.
 *
 * Of course, for this conversion to make sense, the data stored should better
 * be of textual nature.
 *
 * If the conversion is successfull, the inner buffer is replaced by its
 * wide character equivalent.
 *
 * If the fonction fails to convert the data, it will returns -1 and leave the
 * buffer unchanged.
 *
 */

/* -------------------------------------------------------------------------- */

public int string_swap(m_string *string);

/**
 * @ingroup string
 * @fn int string_swap(m_string *string)
 * @param string the string to be swapped
 * @return -1 if an error occured, 0 otherwise.
 *
 * This function reverses the endianness of a binary string, without
 * considerations for the current system endianness.
 *
 * Since it is a quite destructive transformation, all tokens are destroyed.
 *
 */

/* -------------------------------------------------------------------------- */

public inline m_string *string_dup(const m_string *string);

/**
 * @ingroup string
 * @fn m_string *string_dup(const m_string *string)
 * @param string the string to be duplicated.
 * @return a pointer to a new string, or NULL.
 *
 * This function simply makes an exact copy of the m_string given in parameter,
 * and returns it.
 *
 * Please notice that the tokens are *not* copied along with the data, so if
 * a splitted string is copied, its copy has to be explicitely splitted as
 * well.
 *
 * If for whatever reason copying the string is not possible, the function
 * will return NULL.
 *
 */

/* -------------------------------------------------------------------------- */

public inline char *string_dups(const char *string, size_t len);

/**
 * @ingroup string
 * @fn char *string_dups(const char *string, size_t len)
 * @param string the C string to duplicate
 * @param len the length of the C string
 * @return a pointer to a new C string of the given length, or NULL
 *
 * This function copies the first @b len bytes of the given @b string in a
 * freshly allocated buffer.
 *
 * It automatically adds the terminating NUL char at the end of the buffer.
 *
 * The returned buffer should be destroyed with free() after use.
 *
 */

/* -------------------------------------------------------------------------- */

public m_string *string_free(m_string *string);

/**
 * @ingroup string
 * @fn m_string *string_free(m_string *string)
 * @param string the string to be destroyed.
 * @return always NULL.
 *
 * This function will properly clean up and destroy a m_string structure,
 * including its token if there are any.
 *
 * This function always returns NULL, so it could be used to clean a pointer:
 * junk = string_free(junk);
 *
 */

/* -------------------------------------------------------------------------- */

public size_t string_len(const m_string *string);

/**
 * @ingroup string
 * @fn size_t string_len(const m_string *string)
 * @param string the string the size has to be read.
 * @return the size of the string, or (size_t) -1.
 *
 * This function simply returns the size of the given string. It may be
 * preferable to the SIZE, CLEN or WLEN macros since it performs a NULL check.
 *
 * If the size can not be read, the function returns (size_t) -1.
 *
 */

/* -------------------------------------------------------------------------- */

public size_t string_spc(const m_string *string);

/**
 * @ingroup string
 * @fn size_t string_spc(const m_string *string)
 * @param string the string the buffer space has to be read.
 * @return the allocation size of the string, or (size_t) -1.
 *
 * This function simply returns the allocation space consumed by the buffer
 * of the given string.
 *
 * If the buffer space can not be read, the function returns (size_t) -1.
 *
 */

/* -------------------------------------------------------------------------- */

public int string_dim(m_string *string, size_t size);

/**
 * @ingroup string
 * @fn int string_dim(m_string *string, size_t size)
 * @param string the string to resize.
 * @param size the new size to give to the string.
 * @return -1 if the string could not be resized, 0 otherwise.
 *
 * This function forces the resizing of a string. The buffer will be truncated
 * or extended to the given @b size, without care for the inner data.
 *
 * If the buffer can not be resized (wrong size, no sufficient memory or if the
 * string is marked as not resizable), the string is left untouched and the
 * function returns -1.
 *
 * If the buffer size matches the given @b size, no changes are done to the
 * string and the function returns 0.
 *
 * If the buffer is resized and the string had any tokens, the function destroy
 * them.
 *
 */

/* -------------------------------------------------------------------------- */

public inline int string_extend(m_string *string, size_t size);

/**
 * @ingroup string
 * @fn int string_extend(m_string *string, size_t size)
 * @param string the string to be extended.
 * @param size the new size of the string.
 * @return -1 if an error occurs (see @ref string_dim), 0 otherwise.
 *
 * This function is simply a wrapper around @ref string_dim(), which warrants
 * that the buffer will be resized only if its current size is strictly
 * inferior to the one given.
 *
 * If the string is actually resized, all its tokens will be lost.
 *
 */
 

/* -------------------------------------------------------------------------- */

public inline int string_shrink(m_string *string, size_t size);

/**
 * @ingroup string
 * @fn int string_shrink(m_string *string, size_t size)
 * @param string the string to be shrunk.
 * @param size the new size of the string.
 * @return -1 if an error occurs (see @ref string_dim), 0 otherwise.
 *
 * This function is simply a wrapper around @ref string_dim(), which warrants
 * that the buffer will be resized only if its current size is strictly
 * superior to the one given.
 *
 * If the string is actually resized, all its tokens will be lost.
 *
 */

/* -------------------------------------------------------------------------- */

public m_string *string_movs(m_string *to, off_t o, const char *from, size_t l);

/** string_movw is an alias to conveniently use wide char string */
#define string_movw(s, o, w, l) \
        (string_movs((s), (o), (char *) (w), (l) * sizeof(wchar_t)))

/**
 * @ingroup string
 * @fn m_string *string_movs(m_string *to, off_t o, const char *from, size_t l)
 * @param to the destination string.
 * @param o an offset within the destination string bounds
 * @param from the source string.
 * @param l the length of the C source string.
 * @return NULL if an error occurs, or a pointer to the destination string.
 *
 * This function is quite a swiss army knife to manipulate the m_strings.
 *
 * It basically move @b len bytes from the source string to the offset @b o
 * of the destination string @b to given in parameter. The destination
 * string is automatically resized.
 *
 * It handles aliasing (you can pass it the same string as source and
 * destination), and performs all the dirty work of bound checking and
 * size ajustements.
 *
 * If the given destination string is NULL, it will be allocated, thus the
 * return value of this function should always be checked.
 *
 * @ref string_movs() may fail for multiple reasons:
 * - the given source or its size are not correct
 * - resizing the destination string is not possible or allowed
 * - writing to the destination string is not allowed
 * - the destination string is NULL and can not be allocated
 * - the given offset is out of bound
 * However, in all these case, the strings will be left unchanged and the
 * function will returns NULL.
 *
 * If the manipulation is successfull, the function will return a pointer
 * to the updated string and also destroy all tokens it may had.
 *
 * Most of the classic string manipulation functions provided here use
 * @ref string_movs() as a backend.
 *
 */

/* -------------------------------------------------------------------------- */

public inline m_string *string_cats(m_string *to, const char *from, size_t len);

/** string_catw is an alias to conveniently use wide char string */
#define string_catw(s, w, l) \
        (string_cats((s), (char *) (w), (l) * sizeof(wchar_t)))

/**
 * @ingroup string
 * @fn m_string *string_cats(m_string *to, const char *from, size_t len)
 * @param to the destination string.
 * @param from the source C string.
 * @param len the source C string length.
 * @return NULL if an error occurred, 0 otherwise.
 *
 * This function concatenates the source string to the destination string,
 * using @ref string_movs() as backend.
 *
 */

/* -------------------------------------------------------------------------- */

public inline m_string *string_pres(m_string *to, const char *from, size_t len);

/** string_prew is an alias to conveniently use wide char string */
#define string_prew(s, w, l) \
        (string_pres((s), (char *) (w), (l) * sizeof(wchar_t)))

/**
 * @ingroup string
 * @fn m_string *string_pres(m_string *to, const char *from, size_t len)
 * @param to the destination string.
 * @param from the source C string.
 * @param len the source C string length.
 * @return NULL if an error occurred, 0 otherwise.
 *
 * This function prepends the source string to the destination string,
 * using @ref string_movs() as backend.
 *
 */

/* -------------------------------------------------------------------------- */

public inline m_string *string_cpys(m_string *to, const char *from, size_t len);

/** string_cpyw is an alias to conveniently use wide char string */
#define string_cpyw(s, w, l) \
        (string_cpys((s), (char *) (w), (l) * sizeof(wchar_t)))

/**
 * @ingroup string
 * @fn m_string *string_cpys(m_string *to, const char *from, size_t len)
 * @param to the destination string.
 * @param from the source C string.
 * @param len the source C string length.
 * @return NULL if an error occurred, 0 otherwise.
 *
 * This function copies up to @b len bytes from the source string to the
 * destination string, using @ref string_movs() as backend.
 *
 */

/* -------------------------------------------------------------------------- */

public int string_cmps(const m_string *a, const char *b, size_t len);

/** string_cmpw is an alias to conveniently use wide char string */
#define string_cmpw(s, w, l) \
        (string_cmps((s), (char *) (w), (l) * sizeof(wchar_t)))

/**
 * @ingroup string
 * @fn m_string *string_cmps(const m_string *a, const char *b, size_t len)
 * @param a the destination string.
 * @param b the source C string.
 * @param len the source C string length.
 * @return NULL if an error occurred, 0 otherwise.
 *
 * This function compares up to @b len bytes of from the string A with the
 * string B.
 *
 */

/* -------------------------------------------------------------------------- */

public inline m_string *string_mov(m_string *to, off_t off,
                                   const m_string *from, size_t len);

/**
 * @ingroup string
 * @fn m_string *string_mov(m_string *to, off_t off, const m_string *from,
                            size_t len                                    )
 * @param to the destination string.
 * @param off an offset within the destination string bounds.
 * @param from the source string.
 * @param len the length of the data from the source string to use.
 * @return NULL if an error occured, 0 otherwise.
 *
 * This function is simply a wrapper around @ref string_movs(), please see
 * the documentation of @ref string_movs().
 *
 */

/* -------------------------------------------------------------------------- */

public inline m_string *string_cat(m_string *to, const m_string *from);

/**
 * @ingroup string
 * @fn m_string *string_cat(m_string *to, const m_string *from)
 * @param to the destination string.
 * @param from the source string.
 * @return NULL if an error occured, 0 otherwise.
 *
 * This function is simply a wrapper around @ref string_cats(), please see
 * the documentation of @ref string_cats().
 *
 */

/* -------------------------------------------------------------------------- */

public inline m_string *string_pre(m_string *to, const m_string *from);

/**
 * @ingroup string
 * @fn m_string *string_pre(m_string *to, const m_string *from)
 * @param to the destination string.
 * @param from the source string.
 * @return NULL if an error occured, 0 otherwise.
 *
 * This function is simply a wrapper around @ref string_pres(), please see
 * the documentation of @ref string_pres().
 *
 */

/* -------------------------------------------------------------------------- */

public inline m_string *string_cpy(m_string *to, const m_string *from);

/**
 * @ingroup string
 * @fn m_string *string_cpy(m_string *to, const m_string *from)
 * @param to the destination string.
 * @param from the source string.
 * @return NULL if an error occured, 0 otherwise.
 *
 * This function is simply a wrapper around @ref string_cpys(), please see
 * the documentation of @ref string_cpys().
 *
 */

/* -------------------------------------------------------------------------- */

public inline int string_cmp(const m_string *a, const m_string *b);

/**
 * @ingroup string
 * @fn string_cmp(const m_string *a, const m_string *b)
 * @param a the destination string.
 * @param b the source string.
 * @return NULL if an error occured, 0 otherwise.
 *
 * This function is simply a wrapper around @ref string_cmps(), please see
 * the documentation of @ref string_cmps().
 *
 */

/* -------------------------------------------------------------------------- */

public int string_upper(m_string *string);

/**
 * @ingroup string
 * @fn int string_upper(m_string *string)
 * @param string
 * @return -1 if an error occured, 0 otherwise
 *
 * This function simply converts the internal buffer of the given string to
 * upper case. For this to be meaningful, the string should better holds some
 * text.
 *
 * Since the conversion is done in place, no resizing is done and thus
 * existing tokens are preserved.
 *
 * If for whatever reason an error occurs, the function returns -1.
 *
 */

/* -------------------------------------------------------------------------- */

public int string_lower(m_string *string);

/**
 * @ingroup string
 * @fn int string_upper(m_string *string)
 * @param string
 * @return -1 if an error occured, 0 otherwise
 *
 * This function simply converts the internal buffer of the given string to
 * lower case. For this to be meaningful, the string should better holds some
 * text.
 *
 * Since the conversion is done in place, no resizing is done and thus
 * existing tokens are preserved.
 *
 * If for whatever reason an error occurs, the function returns -1.
 *
 */

/* -------------------------------------------------------------------------- */

public off_t string_finds(const m_string *str, size_t o, const char *sub,
                          size_t len                                     );

/** string_findw is an alias to conveniently use wide char string */
#define string_findw(s, w, l) \
        (string_finds((s), (char *) (w), (l) * sizeof(wchar_t)))

/**
 * @ingroup string
 * @fn off_t string_finds(const m_string *str, off_t o, const char *sub,
                          size_t len                                    )
 * @param str the "haystack".
 * @param o an offset within the bounds of the "haystack".
 * @param sub the "needle".
 * @param len the size of the needle.
 * @return -1 if an error occurs, or the position of the needle in the haystack.
 *
 * This functions searches for a substring (needle ;) starting from the offset
 * @b o in the given string @b str. This function use only binary, byte to byte
 * comparisons, so it can be equally used on binary strings.
 *
 * If the substring is not found or an error occurs, this function will
 * return -1. Otherwise, the position of the first byte of the substring in
 * the string is returned.
 *
 */

/* -------------------------------------------------------------------------- */

public inline off_t string_find(const m_string *string, const m_string *sub);

/**
 * @ingroup string
 * @fn off_t string_find(const m_string *string, const m_string *sub)
 * @param string the "haystack".
 * @param sub the "needle".
 * @return -1 if an error occurs, or the position of the needle in the haystack.
 *
 * This function is simply a wrapper around @ref string_finds(), please see
 * the documentation of @ref string_finds().
 *
 */

/* -------------------------------------------------------------------------- */

public int string_splits(m_string *string, const char *pattern, size_t len);

/** string_splitw is an alias to conveniently use wide char string */
#define string_splitw(s, w, l) \
        (string_splits((s), (char *) (w), (l) * sizeof(wchar_t)))

/**
 * @ingroup string
 * @fn string_splits(m_string *string, const char *pattern, size_t len)
 * @param string the string to be split.
 * @param pattern the token delimiter.
 * @param len the size of the delimiter.
 * @return -1 if an error occured, 0 otherwise.
 *
 * This function use @ref string_finds() to split the given @b string, using
 * a delimiter passed in parameter.
 *
 * The pieces of the string are stored in the @ref token field of the string,
 * and their number in the @ref parts field. They can be accessed more
 * conveniently using the @ref PARTS and @ref TOKEN macros.
 *
 * @note It is important to understand that tokens are *not* standalone
 * strings, they merely points to interesting parts of the string which host
 * them, the 'main string'.
 * So, the tokens are internally marked "static" (they can not be resized or
 * modified) to prevent accidental corruption of the main string.
 * However, the tokens can be transparently used as perfectly valid, read-only
 * m_string with the great majority of the functions of the @ref string module.
 * If you want to modify a token, you should copy it in a standalone m_string
 * structure before, with e.g. @ref string_dup().
 *
 * @warning several string manipulation functions destroy the embedded
 * token of a string after doing their work, to avoid corruptions.
 *
 * You can clean up the tokens of a string at any time by calling
 * @ref string_free_token().
 *
 */

/* -------------------------------------------------------------------------- */

public inline int string_split(m_string *string, const m_string *pattern);

/**
 * @ingroup string
 * @fn int string_split(m_string *string, const m_string *pattern)
 * @param string the string to be split.
 * @param pattern the token delimiter.
 * @return -1 if an error occured, 0 otherwise.
 *
 * This function is simply a wrapper around @ref string_splits(), please see
 * the documentation of @ref string_splits().
 *
 */

/* -------------------------------------------------------------------------- */

public m_string *string_reps(m_string *string, const char *search, size_t slen,
                             const char *rep, size_t rlen                     );

/** string_repw is an alias to conveniently use wide char string */
#define string_repw(s, w, l, rw, rl) \
        (string_reps((s), (char *) (w), (l) * sizeof(wchar_t), (char *) (rw), \
         (rl) * sizeof(wchar_t)))

/**
 * @ingroup string
 * @fn m_string *string_reps(m_string *string, const char *search, size_t slen,
                             const char *rep, size_t rlen                      )
 * @param string the string where an expression should be replaced.
 * @param search the expression to be replaced.
 * @param slen the length of this expression.
 * @param rep the replacement string.
 * @param rlen the length of the replacement string.
 * @return NULL if an error occured, a pointer to the main string otherwise.
 *
 * This function searches for the given @b search string inside the main
 * @b string, and replaces each occurence by the provided @b rep string.
 *
 * If a NULL @b rep arguement is given, the target is obviously replaced by...
 * nothing, which means @ref string_reps() may be easily used to destroy a
 * a substring as well.
 *
 * The original string is automatically resized to fit the modifications, but
 * since it is a destructive transformation, all its token will be deleted
 * in order to avoid corruptions.
 *
 */

/* -------------------------------------------------------------------------- */

public m_string *string_rep(m_string *string, const m_string *search,
                            const char *rep, size_t rlen             );

/**
 * @ingroup string
 * @fn m_string *string_rep(m_string *string, const m_string *search,
                            const char *rep, size_t rlen             )
 * @param string the string where the an expression should be replaced.
 * @param search the expression to be replaced.
 * @param rep the replacement string.
 * @param rlen the length of the replacement string.
 * @return NULL if an error occured, a pointer to the main string otherwise.
 *
 * This function is simply a wrapper around @ref string_reps(), please see
 * the documentation of @ref string_reps().
 *
 */

/* -------------------------------------------------------------------------- */

public inline m_string *string_rems(m_string *str, const char *rem, size_t len);

/** string_remw is an alias to conveniently use wide char string */
#define string_remw(s, w, l) \
        (string_rems((s), (char *) (w), (l) * sizeof(wchar_t)))

/**
 * @ingroup string
 * @fn m_string *string_rems(m_string *str, const char *rem, size_t len)
 * @param str the string to be processed
 * @param rem the substring which must be removed
 * @param len the length of the substring
 * @return NULL if an error occured, a pointer to the main string otherwise.
 *
 * This function is simply a wrapper around @ref string_reps() with a NULL
 * replacement string. This way, all occurences of the target substring
 * are dropped.
 *
 * Please see the documentation of @ref string_reps() for further details.
 *
 * Since it uses @ref string_reps() as a backend, this function cause the
 * destruction of all the tokens of the main string.
 *
 */

/* -------------------------------------------------------------------------- */

public inline m_string *string_rem(m_string *string, const m_string *rem);

/**
 * @ingroup string
 * @fn m_string *string_rem(m_string *string, const m_string *rem)
 * @param string the string to be processed
 * @param rem the substring which must be removed
 * @return NULL if an error occured, a pointer to the main string otherwise.
 *
 * This function is simply a wrapper around @ref string_reps() with a NULL
 * replacement string. This way, all occurences of the target substring
 * are dropped.
 *
 * Please see the documentation of @ref string_reps() for further details.
 *
 * Since it uses @ref string_reps() as a backend, this function cause the
 * destruction of all the tokens of the main string.
 *
 */

/* -------------------------------------------------------------------------- */

public m_string *string_b64s(const char *s, size_t size, size_t linesize);

/**
 * @ingroup string
 * @fn m_string *string_b64s(const char *s, size_t size, size_t linesize)
 * @param s the string to be processed
 * @param size the length of the string
 * @param linesize the maximal length of a base64 line
 * @return NULL if an error occured, a pointer to the base64 string otherwise.
 *
 * This function returns a base64 encoded copy of the given string, with
 * @b linesize characters per line.
 *
 * If @b linesize is set to 0, the base64 string will be written as this,
 * without additionnel CRLF.
 *
 * If a @b linesize is provided, it should be a multiple of 4 and not greater
 * than 72, or the function will return NULL.
 *
 */

/* -------------------------------------------------------------------------- */

public inline m_string *string_b64(const m_string *s, size_t linesize);

/**
 * @ingroup string
 * @fn m_string *string_b64(const m_string *s, size_t linesize)
 * @param s the string to be processed
 * @param linesize the maximal length of a base64 line
 * @return NULL if an error occured, a pointer to the base64 string otherwise.
 *
 * This function is simply a wrapper around @ref string_b64s(), please see
 * the documentation of @ref string_b64s().
 *
 */

/* -------------------------------------------------------------------------- */

public m_string *string_deb64s(const char *s, size_t size);

/**
 * @ingroup string
 * @fn m_string *string_deb64s(const char *s, size_t size)
 * @param s the string to be processed
 * @param size the length of the string
 * @return NULL if an error occured, a pointer to the decoded string otherwise.
 *
 * This function decodes a base64 encoded string to plain text, handling
 * eventual embedded CRLFs.
 *
 */

/* -------------------------------------------------------------------------- */

public inline m_string *string_deb64(const m_string *s);

/**
 * @ingroup string
 * @fn m_string *string_deb64(const m_string *s)
 * @param s the string to be processed
 * @return NULL if an error occured, a pointer to decoded the string otherwise.
 *
 * This function is simply a wrapper around @ref string_deb64s(), please see
 * the documentation of @ref string_deb64s().
 *
 */

/* -------------------------------------------------------------------------- */

public void string_free_token(m_string *string);

/**
 * @ingroup string
 * @fn void string_free_token(m_string *string)
 * @param string the string to be cleaned.
 * @return void
 *
 * This function deletes all the tokens of the given string.
 * See @ref m_string or @ref string_splits() for more informations about
 * the tokens.
 *
 */

/* -------------------------------------------------------------------------- */

#endif
