#include "m_port_mmap.h"

/* -------------------------------------------------------------------------- */
#ifdef WIN32 /* WIN32 mmap() compatibility module */
/* -------------------------------------------------------------------------- */

public inline int get_page_size()
{
    #ifndef __WINE__
    SYSTEM_INFO info;
    GetSystemInfo(& info);
    return info.dwPageSize;
    #else
    /* FIXME dwPageSize is set to 0 under winelib ? */
    return 4096;
    #endif
}

/* -------------------------------------------------------------------------- */

public void *mmap(void *start, size_t len, int prot, int flags, int fd,
                  off_t offset                                         )
{
    /*
      This is a minimal implementation of the *NIX mmap syscall for WIN32.
      It supports the following protections flags :
      PROT_NONE, PROT_READ, PROT_WRITE, PROT_EXEC
      It handles the following mapping flags:
      MAP_FIXED, MAP_SHARED, MAP_PRIVATE (POSIX) and MAP_ANONYMOUS (SVID)
    */

    void *ret = MAP_FAILED;
    HANDLE hmap = NULL;
    long wprot = 0, wflags = 0;

    if ( ((fd = _get_osfhandle(fd)) == -1) && (~flags & MAP_ANONYMOUS) ) {
        /* non-file-backed mapping is only allowed with MAP_ANONYMOUS */
        errno = EBADF; return MAP_FAILED;
    }

    /* map *NIX protections and flags to their WIN32 equivalents */
    if ( (prot & PROT_READ) && (~prot & PROT_WRITE) ) {
        /* read only, maybe exec */
        wprot = (prot & PROT_EXEC) ? (PAGE_EXECUTE_READ) : (PAGE_READONLY);
        wflags = (prot & PROT_EXEC) ? (FILE_MAP_EXECUTE) : (FILE_MAP_READ);
    } else if (prot & PROT_WRITE) {
        /* read/write, maybe exec */
        if ( (flags & MAP_SHARED) && (~flags & MAP_PRIVATE) ) {
            /* changes are committed to the file */
            wprot = (prot & PROT_EXEC) ?
                    (PAGE_EXECUTE_READWRITE) : (PAGE_READWRITE);
            wflags = (prot & PROT_EXEC) ?
                    (FILE_MAP_EXECUTE) : (FILE_MAP_WRITE);
        } else if ( (flags & MAP_PRIVATE) && (~flags & MAP_SHARED) ) {
            /* does not affect the original file */
            wprot = PAGE_WRITECOPY; wflags = FILE_MAP_COPY;
        } else {
            /* MAP_PRIVATE + MAP_SHARED is not allowed, abort */
            errno = EINVAL; return MAP_FAILED;
        }
    }

    /* create the windows map object */
    hmap = CreateFileMapping((HANDLE) fd, NULL, wprot, 0, len, NULL);

    if (! hmap) {
        /* the fd was checked before, so it must have bad access rights */
        errno = EPERM; return MAP_FAILED;
    }

    /* create the map itself */
    ret = MapViewOfFileEx(hmap,
                          wflags,
                          0,
                          offset,
                          len,
                          (flags & MAP_FIXED) ? (start) : (NULL)
                         );

    /* Drop the map, it will not be deleted until last 'view' is closed */
    CloseHandle(hmap);

    if (! ret) {
        /* if MAP_FIXED was set, the address was probably wrong */
        errno = (flags & MAP_FIXED) ? (EINVAL) : (ENOMEM); return MAP_FAILED;
    }

    return ret;
}

/* -------------------------------------------------------------------------- */

public int munmap(void *start, unused size_t _dummy)
{
    /*
      This is a minimal implementation of the *NIX munmap syscall for WIN32.
      The size parameter is ignored under Win32.
    */

    if (start == NULL) { errno = EINVAL; return -1; }

    return UnmapViewOfFile(start) ? 0 : -1;
}

/* -------------------------------------------------------------------------- */

public void *shm_alloc(const char *name, size_t size)
{
    HANDLE hmap = NULL;
    void *ret = NULL;

    if (! name || ! size) return NULL;

    /* create a named mapping object */
    hmap = CreateFileMapping(INVALID_HANDLE_VALUE,
                             NULL,
                             PAGE_READWRITE,
                             0,
                             size,
                             name
                            );

    /* we don't want to inadvertently open an existing shared memory object */
    if (GetLastError() == ERROR_ALREADY_EXISTS) { CloseHandle(hmap); return NULL; }

    if (! hmap) { perror(ERR(shm_alloc, CreateFileMapping)); return NULL; }

    /* map the shared memory */
    ret = MapViewOfFileEx(hmap, FILE_MAP_WRITE, 0, 0, size, NULL);

    CloseHandle(hmap);

    if (! ret) { perror(ERR(shm_alloc, MapViewOfFileEx)); return NULL; }

    return ret;
}

/* -------------------------------------------------------------------------- */

public void *shm_attach(const char *name, size_t size)
{
    HANDLE hmap = NULL;
    void *ret = NULL;

    if (! name || ! size) return NULL;

    hmap = OpenFileMapping(PAGE_READWRITE, 0, name);
    if (! hmap) { perror(ERR(shm_alloc, CreateFileMapping)); return NULL; }

    /* map the shared memory */
    ret = MapViewOfFileEx(hmap, FILE_MAP_WRITE, 0, 0, size, NULL);

    CloseHandle(hmap);

    if (! ret) { perror(ERR(shm_alloc, MapViewOfFileEx)); return NULL; }

    return ret;
}

/* -------------------------------------------------------------------------- */

public void shm_detach(void *start, unused size_t _dummy)
{
    /* we already dropped the reference of the map object, just unmap it */
    UnmapViewOfFile(start);
}

/* -------------------------------------------------------------------------- */

public void shm_free(unused const char *_name, void *start, unused size_t _dummy)
{
    /* we already dropped the reference of the map object, just unmap it */
    UnmapViewOfFile(start);
}

/* -------------------------------------------------------------------------- */
#else /* POSIX compliant systems */
/* -------------------------------------------------------------------------- */

public inline int get_page_size(void)
{
    return sysconf(_SC_PAGESIZE);
}

/* -------------------------------------------------------------------------- */

/* mmap(2) and munmap(2) are already there, wrap around shm_open(2) */

public void *shm_alloc(const char *name, size_t size)
{
    int shm = 0;
    void *ret = NULL;

    if (! name || ! size) return NULL;

    shm = shm_open(name, O_RDWR | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR);
    if (shm == -1) { perror(ERR(shm_alloc, shm_open)); return NULL; }

    if (shm && ftruncate(shm, size) == -1) {
        perror(ERR(shm_alloc, ftruncate));
        goto _err_shm;
    }

    /* map the shared memory to the process address space */
    ret = mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_SHARED, shm, 0x0);
    if (ret == MAP_FAILED) { perror(ERR(shm_alloc, mmap)); goto _err_shm; }

    close(shm);

    return ret;

_err_shm:
    close(shm); shm_unlink(name);
    return NULL;
}

/* -------------------------------------------------------------------------- */

public void *shm_attach(const char *name, size_t size)
{
    int shm = 0;
    void *ret = NULL;

    if (! name || ! size) return NULL;

    /* try to get the shared memory descriptor */
    shm = shm_open(name, O_RDWR, 0);
    if (shm == -1) { perror(ERR(shm_attach, shm_open)); }

    /* map the shared memory to the process address space */
    ret = mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_SHARED, shm, 0x0);
    if (ret == MAP_FAILED) { perror(ERR(shm_alloc, mmap)); ret = NULL; }

    close(shm);

    return ret;
}

/* -------------------------------------------------------------------------- */

public inline void shm_detach(void *start, size_t size)
{
    munmap(start, size);
}

/* -------------------------------------------------------------------------- */

public void shm_free(const char *name, void *start, size_t size)
{
    /* unmap and unlink */
    munmap(start, size);
    shm_unlink(name);
}

/* -------------------------------------------------------------------------- */
#endif
/* -------------------------------------------------------------------------- */
