#include "m_port_socket.h"

#ifdef _POSIX_EMULATION
/* this mutex is used to avoid corruption when using inet_ntoa */
static pthread_mutex_t _not_reentrant = PTHREAD_MUTEX_INITIALIZER;
#endif

/* -------------------------------------------------------------------------- */
#ifdef WIN32 /* WIN32 Winsock2 */
/* -------------------------------------------------------------------------- */

/* winsock2 error codes */
static int err[] = {
                     0, WSAEINTR, WSAEBADF, WSAEACCES, WSAEFAULT, WSAEINVAL,
                     WSAEMFILE, WSAEWOULDBLOCK, WSAEINPROGRESS, WSAEALREADY,
                     WSAENOTSOCK, WSAEDESTADDRREQ, WSAEMSGSIZE, WSAEPROTOTYPE,
                     WSAENOPROTOOPT, WSAEPROTONOSUPPORT, WSAESOCKTNOSUPPORT,
                     WSAEOPNOTSUPP, WSAEPFNOSUPPORT, WSAEAFNOSUPPORT,
                     WSAEADDRINUSE, WSAEADDRNOTAVAIL, WSAENETDOWN,
                     WSAENETUNREACH, WSAENETRESET, WSAECONNABORTED, 
                     WSAECONNRESET, WSAENOBUFS, WSAEISCONN, WSAENOTCONN,
                     WSAESHUTDOWN, WSAETOOMANYREFS, WSAETIMEDOUT,
                     WSAECONNREFUSED, WSAELOOP, WSAENAMETOOLONG, WSAEHOSTDOWN,
                     WSAEHOSTUNREACH, WSAENOTEMPTY, WSAEPROCLIM, WSAEUSERS,
                     WSAEDQUOT, WSAESTALE, WSAEREMOTE, WSASYSNOTREADY,
                     WSAVERNOTSUPPORTED, WSANOTINITIALISED, WSAEDISCON,
                     WSAHOST_NOT_FOUND, WSANO_DATA
                   };

/* winsock2 error strings */
static const char *str[] = {
                             "No error", "Interrupted system call",
                             "Bad file number", "Permission denied",
                             "Bad address", "Invalid argument",
                             "Too many open sockets", "Operation would block",
                             "Operation now in progress",
                             "Operation already in progress",
                             "Socket operation on non-socket",
                             "Destination address required",
                             "Message too long",
                             "Protocol wrong type for socket",
                             "Bad protocol option", "Protocol not supported",
                             "Socket type not supported",
                             "Operation not supported on socket",
                             "Protocol family not supported",
                             "Address family not supported",
                             "Address already in use",
                             "Can't assign requested address",
                             "Network is down", "Network is unreachable",
                             "Net connection reset",
                             "Software caused connection abort",
                             "Connection reset by peer",
                             "No buffer space available",
                             "Socket is already connected",
                             "Socket is not connected",
                             "Can't send after socket shutdown",
                             "Too many references, can't splice"
                             "Connection timed out",
                             "Connection refused",
                             "Too many levels of symbolic links",
                             "File name too long", "Host is down",
                             "No route to host", "Directory not empty",
                             "Too many users", "Disc quota exceeded",
                             "Stale NFS file handle",
                             "Too many level of remote in path",
                             "Network system is unavailable",
                             "Winsock version out of range",
                             "WSAStartup not yet called",
                             "Graceful shutdown in progress",
                             "Host not found",
                             "No host data of that type was found"
                           };



/* -------------------------------------------------------------------------- */

private const char *_socket_win32_strerror(void)
{
    unsigned int i = 0;

    for (i = 0; i < sizeof(err); i ++)
        if (errno == err[i]) return str[i];

    return NULL;
}

/* -------------------------------------------------------------------------- */

public size_t socket_sendfile(int out, int in, off_t *off, size_t len)
{
    HANDLE handle = INVALID_FILE_HANDLE;

    if (! off) { return -1; errno = EINVAL; }

    /* try seek to the given offset */
    if (_lseek(in, *off, SEEK_SET) == -1) { errno = EINVAL; return -1; }

    /* get the windows API file handle */
    if ( (handle = (HANDLE) _get_osfhandle(in)) == INVALID_FILE_HANDLE) {
        errno = EBADF; return -1;
    }

    /* don't use the TCP_CORK-like feature, since TransmitFile() is blocking;
       so at least we could suspend writing if the header or footer blocks */
    if (! TransmitFile(out, handle, len, 0, NULL, NULL, 0)) {
        errno = EIO; return -1;
    }

    *off += len;

    return len;
}

/* -------------------------------------------------------------------------- */
#elif defined (__FreeBSD__)
/* -------------------------------------------------------------------------- */

public size_t socket_sendfile(int out, int in, off_t *off, size_t len)
{
    off_t written = 0;
    int ret = -1;

    if ( ! off || *off < 0) return -1;
    
    ret = sendfile(in, out, *off, len, NULL, & written, 0);

    if (ret == 0) { ret = written; *off += written; }

    return ret;
}

/* -------------------------------------------------------------------------- */
#elif defined (__sun)
/* -------------------------------------------------------------------------- */

public size_t socket_sendfile(int out, int in, off_t *off, size_t len)
{
    size_t written = 0;
    int ret = 0;
    struct sendfilevec vector;

    memset(& vector, 0, sizeof(vector));
    vector.sfv_fd = in;
    vector.sfv_off = *off;
    vector.sfv_len = len;

    ret = sendfilev(out, & vector, 1, & written);

    if (ret == 0) { ret = written; *off += written; }

    return ret;
}

/* -------------------------------------------------------------------------- */
#else
/* -------------------------------------------------------------------------- */

/* generic implementation */
#ifndef __linux__
public size_t socket_sendfile(int out, int in, off_t *off, size_t len)
{
    size_t byte_read = 0, byte_written = 0, written = 0;
    unsigned char *buffer;

    if ( ! off || *off < 0) return -1;

    if ( (buffer = malloc(len * sizeof(*buffer))) == NULL) {
        perror(ERR(_socket_sendfile, malloc)); return -1;
    }

    /* to be portable, 'in' should not be a socket */
    while ( (byte_read = read(in, buffer, len)) == -1) {
        /* check the reason of the error */
        if (errno == EINTR) continue; else break;
    }

    /* abort right here if a read error occured */
    if (byte_read == -1 || byte_read == 0) { free(buffer); return -1; }

    /* FIXME: check if the out descriptor is a socket or a file */

    /* ensure all that was read is sent, even if this means blocking a bit */
    while (1) {
        byte_written = write(out, buffer + written, byte_read);

        if (byte_written == -1 || byte_written == 0) break;

        written += byte_written; byte_read -= byte_written;

        if (byte_read == 0) break;
    }

    free(buffer);

    return written;
}
#endif

/* -------------------------------------------------------------------------- */
#endif
/* -------------------------------------------------------------------------- */

/*
  This is a small compatibility layer provided for OS or compilers which do not
  implement the POSIX protocol independant network functions. The assumption is
  made that such an OS/compiler probably does not either properly implement
  IPv6, so only the classic BSD IPv4 API is used for a better portability.
*/

/* -------------------------------------------------------------------------- */
#ifdef _POSIX_EMULATION
/* -------------------------------------------------------------------------- */

/* define the addrinfo structure */
struct addrinfo {
    int     ai_flags;
    int     ai_family;
    int     ai_socktype;
    int     ai_protocol;
    size_t  ai_addrlen;
    struct sockaddr *ai_addr;
    char   *ai_canonname;
    struct addrinfo *ai_next;
};

/* use the classic sockaddr_in */
#define sockaddr_storage sockaddr_in

/* use the minimalist implementation of the POSIX functions */
#undef getaddrinfo
#undef getnameinfo
#undef freeaddrinfo
#define getaddrinfo  _getaddrinfo
#define getnameinfo  _getnameinfo
#define freeaddrinfo _freeaddrinfo

/* remove gai_strerror */
#undef gai_strerror
#define gai_strerror(i) "unknown error."

/* netdb.h definitions */
#define AI_PASSIVE     0x0001  /* Socket address is intended for `bind'.  */
#define AI_NUMERICHOST 0x0004  /* Don't use name resolution.  */
#define EAI_NONAME       -2    /* NAME or SERVICE is unknown.  */
#define EAI_FAIL         -4    /* Non-recoverable failure in name res.  */
#define EAI_FAMILY       -6    /* `ai_family' not supported.  */
#define EAI_MEMORY       -10   /* Memory allocation failure.  */
#define NI_NUMERICHOST 1       /* Don't try to look up hostname.  */
#define NI_NUMERICSERV 2       /* Don't convert port number to name.  */

/* default size for the host and service buffers */
#define NI_MAXHOST      1025
#define NI_MAXSERV      32

/* -------------------------------------------------------------------------- */

public int getaddrinfo(const char *node, const char *service,
                        const struct addrinfo *hints, struct addrinfo **res)
{
    /** @brief *Very* minimal getaddrinfo() implementation */
    /*
      This is a *very* minimal implementation of getaddrinfo(), to be able
      to compile and run with an OS/compiler which is not POSIX compliant.
      So, this small implementation only meets the needs of MammouthServer
      and use the most common BSD functions.
      This "implementation" only supports IPv4.
    */

    struct addrinfo *ai = NULL;
    int len = sizeof(struct sockaddr_in);
    struct sockaddr_in *sa = NULL;

    /* not a full implementation, bail out if no hint was given */
    if (! res || ! hints) return EAI_FAIL;

    /* only support AI_NUMERICHOST */
    if (! hints->ai_flags & AI_NUMERICHOST) return EAI_FAIL;

    /* only support IPv4 */
    if (hints->ai_family != AF_INET) return EAI_FAMILY;

    /* exit conforming to the specs if no node nor service was given */
    if (! node && ! service) return EAI_NONAME;

    if ( ! (ai = malloc(sizeof(*ai))) ) return EAI_MEMORY;
    ai->ai_family   = hints->ai_family; ai->ai_socktype = hints->ai_socktype;
    ai->ai_protocol = hints->ai_protocol; ai->ai_addrlen = len;

    if (! (ai->ai_addr = malloc(len)) ) { free(ai); return EAI_MEMORY; }
    sa = (void *) ai->ai_addr; memset(sa, 0, len);

    sa->sin_family = AF_INET;
    sa->sin_addr.s_addr = (hints->ai_flags & AI_PASSIVE) ?
                          (0x0) : (node ? inet_addr(node) : -1);
    if (service) sa->sin_port = htons((short) atoi(service));

    /* all went fine */
    ai->ai_next = NULL; *res = ai;

    return 0;
}

/* -------------------------------------------------------------------------- */

public int getnameinfo(const struct sockaddr *s, int salen, char *host,
                        size_t hostlen, char *serv, size_t servlen, int flags)
{
    /** @brief *Very* minimal getnameinfo() implementation */
    /*
      This is a *very* minimal implementation of getnameinfo(), to be able
      to compile and run with an OS/compiler which is not POSIX compliant.
      So, this small implementation only meets the needs of MammouthServer
      and use the most common BSD functions.
      This "implementation" only supports IPv4.
    */

    const char *ip = NULL;
    struct sockaddr_in *sa = (void *) s;

    /* neither hostname nor service name were requested */
    if (! host && ! serv) return EAI_NONAME;

    /* wrong address */
    if (! sa || salen != sizeof(*sa)) return EAI_FAMILY;

    /* minimal implementation, only support NI_NUMERICHOST & NI_NUMERICSERV */
    if (host && (flags & NI_NUMERICHOST) ) {
        pthread_mutex_lock(_not_reentrant);
        if ((ip = inet_ntoa(sa->sin_addr))) strncpy(host, ip, hostlen - 1);
        pthread_mutex_unlock(_not_reentrant);
    } else
        return EAI_FAIL;

    if (serv && (flags & NI_NUMERICSERV) )
        snprintf(serv, servlen - 1, "%i", ntohs(sa->sin_port));
    else
        return EAI_FAIL;

    /* all went fine */
    return 0;   
}

/* -------------------------------------------------------------------------- */

public void freeaddrinfo(struct addrinfo *res)
{
    /** @brief *Very* minimal freeaddrinfo() implementation */
    /*
      This is a *very* minimal implementation of freeaddrinfo(), to be able
      to compile and run with an OS/compiler which is not POSIX compliant.
      So, this small implementation only meets the needs of MammouthServer
      and use the most common BSD functions.
      This "implementation" only supports IPv4.
    */

    if (res && res->ai_addr) free(res->ai_addr); free(res);
}

/* -------------------------------------------------------------------------- */
#endif
/* -------------------------------------------------------------------------- */

/*
 * The following implements sendfd() and recvfd() to send or receive sockets
 * over Unix domain sockets.
 *
 */

/* -------------------------------------------------------------------------- */
#ifdef WIN32
/* -------------------------------------------------------------------------- */

public int sendfd(int sock_fd, int send_me_fd)
{
    WSAPROTOCOL_INFO info;
    DWORD pid = 0;
    size_t len = 0;

    /* the receiver must give us a pid */
    if (recv(sock_fd, (char *) & pid, sizeof(pid), 0x0) == -1) {
        serror(ERR(sendfd, recv)); return -1;
    }

    if (WSADuplicateSocket(send_me_fd, pid, & info) == -1) {
        serror(ERR(sendfd, WSADuplicateSocket)); return -1;
    }

    /* send the raw struct over the wire */
    if (send(sock_fd, (char *) & info, sizeof(info), 0x0) == -1) {
        serror(ERR(sendfd, send)); return -1;
    }

    return 0;
}

/* -------------------------------------------------------------------------- */

public int recvfd(int sock_fd)
{
    WSAPROTOCOL_INFO info;
    DWORD pid = _getpid();
    int ret = 0;
    
    if (send(sock_fd, (char *) & pid, sizeof(pid), 0x0) == -1) {
        serror(ERR(recvfd, send)); return -1;
    }

    if (recv(sock_fd, (char *) & info, sizeof(info), 0x0) == -1) {
        serror(ERR(recvfd, recv)); return -1;
    }

    #define _FPI FROM_PROTOCOL_INFO
    if ( (ret = WSASocket(_FPI, _FPI, _FPI, & info, 0, 0x0)) == -1) {
        serror(ERR(recvfd, WSASocket)); return -1;
    }
    #undef _FPI

    return ret;
}

/* -------------------------------------------------------------------------- */
#else
/* -------------------------------------------------------------------------- */

/*
 * Copyright (c) 2000 Sampo Kellomaki <sampo@iki.fi>, All Rights Reserved.
 * This module may be copied under the same terms as the perl itself.
 *
 */

/* I test here for __sun for lack of anything better, but I
 * mean Solaris 2.6. The idea of undefining SCM_RIGHTS is
 * to force the headers to behave BSD 4.3 way which I have
 * tested to work.
 *
 * In general, if you have compilation errors, you might consider
 * adding a test for your platform here.
 */
#if defined(__sun)
#undef SCM_RIGHTS
#endif

#ifdef SCM_RIGHTS

/* It seems various versions of glibc headers (i.e.
 * /usr/include/socketbits.h) miss one or more of these */

#ifndef CMSG_DATA
# define CMSG_DATA(cmsg) ((cmsg)->cmsg_data)
#endif

#ifndef CMSG_NXTHDR
# define CMSG_NXTHDR(mhdr, cmsg) __cmsg_nxthdr (mhdr, cmsg)
#endif

#ifndef CMSG_FIRSTHDR
# define CMSG_FIRSTHDR(mhdr) \
  ((size_t) (mhdr)->msg_controllen >= sizeof (struct cmsghdr)          \
   ? (struct cmsghdr *) (mhdr)->msg_control : (struct cmsghdr *) NULL)
#endif

#ifndef CMSG_ALIGN
# define CMSG_ALIGN(len) (((len) + sizeof (size_t) - 1) \
             & ~(sizeof (size_t) - 1))
#endif

#ifndef CMSG_SPACE
# define CMSG_SPACE(len) (CMSG_ALIGN (len) \
             + CMSG_ALIGN (sizeof (struct cmsghdr)))
#endif

#ifndef CMSG_LEN
# define CMSG_LEN(len)   (CMSG_ALIGN (sizeof (struct cmsghdr)) + (len))
#endif

union fdmsg {
    struct cmsghdr h;
    char buf[CMSG_SPACE(sizeof(int))];
};
#endif

public int sendfd(int sock_fd, int send_me_fd)
{
    int ret = 0;
    struct iovec  iov[1];
    struct msghdr msg;
    
    iov[0].iov_base = &ret;  /* Don't send any data. Note: der Mouse
                              * <mouse@Rodents.Montreal.QC.CA> says
                              * that might work better if at least one
                              * byte is sent. */
    iov[0].iov_len  = 1;
    
    msg.msg_iov     = iov;
    msg.msg_iovlen  = 1;
    msg.msg_name    = 0;
    msg.msg_namelen = 0;

    {
#ifdef SCM_RIGHTS
        /* New BSD 4.4 way (ouch, why does this have to be
         * so convoluted). */

        union  fdmsg  cmsg;
        struct cmsghdr* h;

        msg.msg_control = cmsg.buf;
        msg.msg_controllen = sizeof(union fdmsg);
        msg.msg_flags = 0;
        
        h = CMSG_FIRSTHDR(&msg);
        h->cmsg_len   = CMSG_LEN(sizeof(int));
        h->cmsg_level = SOL_SOCKET;
        h->cmsg_type  = SCM_RIGHTS;
        *((int*)CMSG_DATA(h)) = send_me_fd;
#else
        /* Old BSD 4.3 way. Not tested. */
        msg.msg_accrights = &send_me_fd;
        msg.msg_accrightslen = sizeof(send_me_fd);
#endif  

        if (sendmsg(sock_fd, &msg, 0) < 0) {
            ret = 0;
        } else {
            ret = 1;
        }
    }
    /*fprintf(stderr,"send %d %d %d %d\n",sock_fd, send_me_fd, ret, errno);*/
    return ret;
}

/* -------------------------------------------------------------------------- */

public int recvfd(int sock_fd)
{
    int count;
    int ret = 0;
    struct iovec  iov[1];
    struct msghdr msg;
    
    iov[0].iov_base = &ret;  /* don't receive any data */
    iov[0].iov_len  = 1;
    
    msg.msg_iov = iov;
    msg.msg_iovlen = 1;
    msg.msg_name = NULL;
    msg.msg_namelen = 0;

    {
#ifdef SCM_RIGHTS
        union fdmsg  cmsg;
        struct cmsghdr* h;

        msg.msg_control = cmsg.buf;
        msg.msg_controllen = sizeof(union fdmsg);
        msg.msg_flags = 0;
        
        h = CMSG_FIRSTHDR(&msg);
        h->cmsg_len   = CMSG_LEN(sizeof(int));
        h->cmsg_level = SOL_SOCKET;  /* Linux does not set these */
        h->cmsg_type  = SCM_RIGHTS;  /* upon return */
        *((int*)CMSG_DATA(h)) = -1;
        
        if ((count = recvmsg(sock_fd, &msg, 0)) < 0) {
            ret = 0;
        } else {
            h = CMSG_FIRSTHDR(&msg);   /* can realloc? */
            if (   h == NULL
                || h->cmsg_len    != CMSG_LEN(sizeof(int))
                || h->cmsg_level  != SOL_SOCKET
                || h->cmsg_type   != SCM_RIGHTS ) {
                /* This should really never happen */
                if (h)
                  fprintf(stderr,
                    "%s:%d: protocol failure: %d %d %d\n",
                    __FILE__, __LINE__,
                    h->cmsg_len,
                    h->cmsg_level, h->cmsg_type);
                else
                  fprintf(stderr,
                    "%s:%d: protocol failure: NULL cmsghdr*\n",
                    __FILE__, __LINE__);
                ret = 0;
            } else {
                ret = *((int*)CMSG_DATA(h));
                /*fprintf(stderr, "recv ok %d\n", ret);*/
            }
        }
#else
        int receive_fd;
        msg.msg_accrights = &receive_fd;
        msg.msg_accrightslen = sizeof(receive_fd);

        if (recvmsg(sock_fd, &msg, 0) < 0) {
            ret = 0;
        } else {
            ret = receive_fd;
        }
#endif
    }
    
    /*fprintf(stderr, "recv %d %d %d %d\n",sock_fd, ret, errno, count);*/
    return ret;
}

/* -------------------------------------------------------------------------- */
#endif
/* -------------------------------------------------------------------------- */
